"""
Introduction to TriFusion's process module
==========================================

The `process` subpackage is the main backend of TriFusion's Process and
Statistics modules and of TriSeq and TriStats CLI programs. The most
important classes are defined in the :mod:`~trifusion.process.sequence` module:
:class:`~trifusion.process.sequence.Alignment` and
:class:`~trifusion.process.sequence.AlignmentList`.

What it does
------------

The `process` module contains the classes and functions responsible for
parsing, modifying, writing and plotting alignment data.

Submodules description
----------------------

:mod:`~trifusion.process.base`
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Contains several methods and function that are inherited or used by
:class:`~trifusion.process.sequence.Alignment` and
:class:`~trifusion.process.sequence.AlignmentList` objects, as well as by
the TriSeq and TriStats CLI programs.

:mod:`~trifusion.process.data`
~~~~
Contains the :class:`~trifusion.process.data.Partitions`  class, used by
:class:`~trifusion.process.sequence.Alignment` and
:class:`~trifusion.process.sequence.AlignmentList`
classes to handle partitions in the alignments.

:mod:`~trifusion.process.error_handling`
~~~~~~~~~~~~~~
Contains custom made Exception sub-classes.

:mod:`~trifusion.process.sequence`
~~~~~~~~
Contains the :class:`~trifusion.process.sequence.Alignment`  and
:class:`~trifusion.process.sequence.AlignmentList` classes, responsible
for the majority of the heavy lifting when dealing with alignment files. See
the module's documentation for further details.
"""


#!/usr/bin/python2
# -*- coding: utf-8 -*-

import os
import re

try:
    from process.error_handling import KillByUser
except ImportError:
    from trifusion.process.error_handling import KillByUser


def orthomcl_filter_fasta(input_dir, min_length, max_stop_percent, db, dest,
                             nm=None):

    def handle_seq(seq, length, stop_cnt):
        is_bad = 0
        stop_percent = ((length - stop_cnt) / length) * 100

        if length < min_length or stop_percent > max_stop_percent:
            bad.write(seq + "\n")
            is_bad = 1
        else:
            good.write(seq + "\n")

        return is_bad

    good = open(os.path.join(dest, "backstage_files", db), "w")
    bad = open(os.path.join(dest, "backstage_files", "poorProteins.txt"), "w")

    filenames = [os.path.join(input_dir, x) for x in os.listdir(input_dir)]

    reject_rates = []

    # Setup progression information
    if nm:
        if nm.stop:
            raise KillByUser("")
        nm.total = len(filenames)
        nm.counter = 0

    for filename in filenames:

        if nm:
            if nm.stop:
                raise KillByUser("")
            nm.counter += 1
            nm.msg = "Filtering file {}".format(os.path.basename(filename))

        if filename.startswith('.'):
            continue

        input_file = open(filename, 'r')
        seq_count = 0
        reject_seq_count = 0
        current_seq = ""
        current_len = 0
        current_stop_cnt = 0

        # process lines of one file
        for line in input_file:

            if nm:
                if nm.stop:
                    raise KillByUser("")

            if line.startswith('>'):
                if current_seq:
                    seq_count += 1
                    reject_seq_count += handle_seq(current_seq,
                                                   current_len,
                                                   current_stop_cnt)
                    current_seq = ""
                    current_len = 0
                    current_stop_cnt = 0
            else:
                line_len = len(line)
                current_len += line_len
                line = re.sub('[^A-Za-z]', '', line)
                current_stop_cnt += line_len - len(line)

            current_seq += line

        reject_seq_count += handle_seq(current_seq,
                                       current_len,
                                       current_stop_cnt)
        seq_count += 1

        # add file stats to reject count if it qualifies
        if reject_seq_count:
            pct = reject_seq_count / seq_count * 100
            if pct > 10:
                reject_rates.append([input_file, pct])

        input_file.close()

    good.close()
    bad.close()


__author__ = "Fernando Alves"




#!/usr/bin/env python2
# -*- coding: utf-8 -*-
#
#  Copyright 2012 Unknown <diogo@arch>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import re
from os.path import basename, splitext, join
from os import sep
from collections import OrderedDict


class PartitionException(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class InvalidPartitionFile(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class Partitions(object):
    """Alignment partitions interface for `Alignment` and `AlignmentList`.

    The Partitions class is used to define partitions for `Alignment`
    and `AlignmentList` objects and associate substitution models to
    each partition. After instantiating, partitions may be set in two ways:

      - Partition files: Being Nexus charset blocks and RAxML partition files
        currently supported
      - Tuple-like objects: Containing the ranges and names of the partitions

    Attributes
    ----------
    partition_length : int
        Length of the total partitions.
    partitions : OrderedDict
        Storage of partition names (key) and their range (values).
    partitions_index : list
        The index (starting point) for each partition, including codon
        partitions.
    partitions_alignments : OrderedDict
        Storage of the partition names (key) and their corresponding
        alignment files (values).
    alignments_range : OrderedDict
        Storage of the alignment names (key) and their range (values).
    models : OrderedDict
        Storage of partition names (key) and their models (values).
    merged_files : dict
        Storage of the original range (values) of every alignment file (key).
    counter : int
        Indicator of where the last partition ended.
    partition_format : str
        Format of the original partition file, if any.
    """

    _models = {"mrbayes": {}}

    # =========================================================================
    #   MrBayes models
    # =========================================================================
    """
    MrBayes substitution models are stored in the dictionary
    _models["mrbayes"]. The keys of the dictionary are the name of the
    substitution models (usually in capital letters) and the values will
    contain the instructions to specific such model in a list. Each element
    of the list corresponds to one line
    """

    # GTR
    _models["mrbayes"]["GTR"] = ["nst=6", "statefreqpr=dirichlet(1,1,1,1)"]

    # SYM
    _models["mrbayes"]["SYM"] = ["nst=6", "statefreqpr=fixed(equal)"]

    # HKY
    _models["mrbayes"]["HKY"] = ["nst=2", "statefreqpr=dirichlet(1,1,1,1)"]

    # K2P
    _models["mrbayes"]["K2P"] = ["nst=2", "statefreqpr=fixed(equal)"]

    # F81
    _models["mrbayes"]["F81"] = ["nst=1", "statefreqpr=dirichlet(1,1,1,1)"]

    # JC
    _models["mrbayes"]["JC"] = ["nst=1", "statefreqpr=fixed(equal)"]

    def __init__(self):

        self.partition_length = 0
        """
        The length of the locus may be necessary when partitions are defined
        in the input files using the "." notation, meaning the entire locus.
        Therefore, to convert this notation into workable integers, the size
        of the locus must be provided using the set_length method.
        """

        self.partitions = OrderedDict()
        """
        partitions will contain the name and range of the partitions for a given
        alignment object. Both gene and codon partitions will be stored in this
        attribute, but gene partitions are the main entries. An example of
        different stored partitions is::

            partitions = {"partitionA": ((0, 856), False),
                          "partitionB": ((857, 1450), [857,858,859] }

        "partitionA" is a simple gene partition ranging from 0 to 856, while
        "partitionB" is an assembly of codon partitions. The third element of
        the tuple is destined to codon partitions. If there are none, it should
        be False. If there are codon partitions, a list should be provided with
        the desired initial codons. In the example above, "partitionB" has
        actually 3 partitions starting at the first, second and third sequence
        nucleotide of the main partition.
        """

        self.partitions_index = []
        """
        partitions_index will remember the index of all added partitions. This
        attribute was created because codon models are added to the same parent
        partitions, thus losing their actual index. This is important for
        Nexus files, where models are applied to the index of the partition.
        This will simply store the partition names, which can be accessed using
        their index, or searched to return their index. To better support codon
        partitions, each entry in the partitions_index will consist in a list,
        in which the first element is the partition name, and the second element
        is the index of the subpartition. An example would be::

            partitions_index = [["partA", 0], ["partA", 1], ["partA", 2],
                                ["partB", 0]]

        in which, partA has 3 codon partitions, and partB has only one partition

        """

        self.partitions_alignments = OrderedDict()
        """
        The partitions_alignments attribute will associate the partition with
        the corresponding alignment files. For single alignment partitions,
        this will provide information on the file name. For multiple
        alignments, besides the information of the file names, it will
        associate which alignments are contained in a given partition and
        support multi alignment partitions. An example would be::

            partitions_alignments = {"PartitionA": ["FileA.fas"],
                                     "PartitionB": ["FileB.fas", "FileC.fas"]}

        """

        self.alignments_range = OrderedDict()
        """
        """

        self.models = OrderedDict()
        """
        The self.models attribute will contain the same key list as
        self.partitions and will associate the substitution models to each
        partitions. For each partition, the format should be as follows::

            models["partA"] = [[[..model_params..]],[..model_names..],
                               ["12", "3"]]

        The first element is a list that may contain the substitution model
        parameters for up to three subpartitions, the second element is also
        a list with the corresponding names of the substitution models and
        the third list will store any links between models
        """

        self.merged_files = {}
        """
        This attribute will keep a record of the original ranges of every file
        that was merged. This is useful to split partitions according to files
        or to undo any changes. Each entry should be::

            {"alignment_file1": (0, 1234), "alignment_file2": (3444, 6291)}
        """

        self.counter = 0
        """
        The counter attribute will be used as an indication of where the last
        partition ends when one or more partitions are added
        """

        self.partition_format = None

    def __iter__(self):
        """Iterator behavior for `Partitions`.

        The class iterator will iterate over a list containing the partition
        names and a modified version of their ranges that is compatible with
        other software (unlike the 0 offset of python)

        Returns
        _ : iter
            Iterator of `partitions.items()`.
        """

        return iter(self.partitions.items())

    def reset(self, keep_alignments_range=False):
        """Clears partitions and attributes

        Clears partitions and resets object to __init__ state. The original
        alignment range can be retained by setting the `keep_alignments_range`
        argument to True.

        Parameters
        ----------
        keep_alignments_range : bool
            If True, the `alignments_range` attribute will not be reset.
        """

        self.partitions = OrderedDict()
        self.partitions_index = []
        self.partitions_alignments = OrderedDict()
        self.models = OrderedDict()
        self.counter = 0
        if not keep_alignments_range:
            self.alignments_range = OrderedDict()

    def iter_files(self):
        """Iterates over `partitions_alignments.items()`.

        Returns
        -------
        _ : iter
            Iterator of `partitions_alignments.items()`.
        """

        return iter(self.partitions_alignments.items())

    def set_length(self, length):
        """Set total length of current locus (over all partitions).

        Sets the length of the locus. This may be important to convert certain
        partition defining nomenclature, such as using the "." to indicate
        whole length of the alignment

        Parameters
        ----------
        length : int
            Integer that will be set as `partition_length`.
        """

        self.partition_length = length

    #===========================================================================
    # Parsers
    #===========================================================================

    @staticmethod
    def _get_file_format(partition_file):
        """Guesses the format of the partition file (Nexus or RAxML's).

        Returns
        -------
        partition_format : str
            Format of the partition file ("nexus" or "raxml").
        """

        file_handle = open(partition_file)

        # Skips first empty lines, if any
        header = file_handle.readline()
        while header.startswith("\n"):
            header = next(file_handle)

        fields = header.split()
        if fields[0].lower() == "charset":
            partition_format = "nexus"
        else:
            partition_format = "raxml"

        return partition_format

    def read_from_file(self, partitions_file):
        """Parses partitions from file

        This method parses a file containing partitions. It supports
        partitions files similar to RAxML's and NEXUS charset blocks. The
        NEXUS file, however, must only contain the charset block. The
        model_nexus argument provides a namespace for the model variable in
        the nexus format, since this information is not present in the file.
        However, it assures consistency on the Partition object.

        Parameters
        ----------
        partitions_file : str
            Path to partitions file.

        Raises
        ------
        PartitionException
            When one partition definition cannot be parsed.
        """

        # Resets previous partitions (except alignments_range)
        self.reset(keep_alignments_range=True)

        # Get the format of the partition file
        self.partition_format = self._get_file_format(partitions_file)

        part_file = open(partitions_file)

        # In order to support unsorted partition ranges, the complete
        # partition set will be stored temporary in memory. Even very large
        # partition files should result in relatively small data structures.
        # Once this variable is populated, it will be sorted according to the
        # first element of the range.
        temp_ranges = []

        # TODO: Add support for codon partitions in raxml format
        if self.partition_format == "raxml":
            for p, line in enumerate(part_file):

                # Ignore empty lines
                if line.strip() == "":
                    continue

                # A wrongly formatted raxml partition file may be provided, in
                # which case an IndexError exception will be raised. This will
                # handle that exception
                try:
                    fields = line.split(",")
                    # Get partition name as string
                    partition_name = fields[1].split("=")[0].strip()
                    # Get partition range as list of int
                    partition_range_temp = fields[1].split("=")[1]
                    try:
                        partition_range = [
                            int(x) - 1 for x in
                            partition_range_temp.strip().split("-")]

                    except ValueError as e:
                        # A ValueError may be raise when there is a "."
                        # notation in the partition range. If so, convert
                        # the "." to the sequence lenght. If no sequence lenght
                        # has been provided raise another exception
                        pr = partition_range_temp.strip().split("-")
                        if pr[1] == ".":
                            if self.partition_length:
                                partition_range = [int(pr[0]) - 1,
                                                   self.partition_length - 1]
                            else:
                                raise PartitionException(
                                    "The length of the locus must be "
                                    "provided when partitions are "
                                    "defined using '.' notation to "
                                    "mean full length")
                        else:
                            raise e

                    # Check which alignment file contains the current partition
                    if self.alignments_range:
                        try:
                            file_name = \
                                [x for x, y in self.alignments_range.items() if
                                 y[0] <= partition_range[0] < y[1]][0]
                        except IndexError:
                            file_name = None
                    else:
                        file_name = None

                    temp_ranges.append([partition_name, file_name,
                                        partition_range])

                except (IndexError, ValueError):
                    return InvalidPartitionFile(
                        "Badly formatted partitions file in line {} "
                        "with:\n\n{}".format(p + 1, line))

        elif self.partition_format == "nexus":
            for line in part_file:
                # Ignore empty lines
                if line.strip() != "":
                    try:
                        res = self.read_from_nexus_string(line,
                                                          return_res=True)
                    except PartitionException as e:
                        return e
                    if res:
                        temp_ranges.append(res)

        # Sort partition ranges according to the first element of the range
        temp_ranges.sort(key=lambda part: part[2][0])

        for name, file_name, part_range in temp_ranges:
            # Add information to partitions storage
            try:
                self.add_partition(name,
                                   locus_range=part_range,
                                   file_name=file_name)
            except InvalidPartitionFile as e:
                return e

    def read_from_nexus_string(self, nx_string, file_name=None,
                               return_res=False):
        """Parses a single nexus string with partition definition.

        Parameters
        ----------
        nx_string : str
            String with partition definition
        file_name : str, optional
            String with name of the file corresponding to the partition.
        return_res : bool
            If True, it will only return the parsed partition information.
            If False, it will add the parsed partition to the `Partitions`
            object.
        """

        try:
            fields = nx_string.split("=")
            partition_name = fields[0].split()[1].strip()

            # If this list has 2 elements, it should be a simple gene partition
            # If it has 3 elements, it should be a codon partition
            partition_full = re.split(r"[-\\]", fields[1].strip().
                                      replace(";", "").replace("/", "\\"))

            # If partition is defined using "." notation to mean full length
            if partition_full[1] == ".":
                if self.partition_length:
                    partition_range = [int(partition_full[0]) - 1,
                                       self.partition_length - 1]
                else:
                    raise PartitionException("The length of the locus must be "
                                             "provided when partitions are "
                                             "defined using '.' notation to "
                                             "mean full length")
            else:
                partition_range = [int(partition_full[0]) - 1,
                                   int(partition_full[1]) - 1]

            # Check which alignment file contains the current partition
            if self.alignments_range:
                try:
                    file_name = [x for x, y in self.alignments_range.items() if
                                 partition_range[0] in xrange(*y) and
                                 partition_range[1] in xrange(*y)][0]
                except IndexError:
                    pass

            if return_res:
                return [partition_name, file_name, partition_range]
            else:
                self.add_partition(partition_name, locus_range=partition_range,
                                   file_name=file_name)
        # If, for some reason, the current line cannot be interpreted as a
        # charset line, ignore it.
        except IndexError:
            if return_res:
                return None
            else:
                pass

    def get_partition_names(self):
        """Returns a list with the name of the partitions

        Returns
        -------
        names : list
            List with names of the partitions. When a parent
            partition has multiple codon partitions, it returns a partition
            name for every codon starting position present.
        """

        names = []

        for part, vals in self.partitions.items():
            if vals[1]:
                names.extend([part + "_%s" % (x + 1) for x in vals[1]])
            else:
                names.append(part)

        return names

    def read_from_dict(self, dict_obj):
        """Reads partition information from a dict object

        Parses partitions defined and stored in a special OrderedDict. The
        values of dict_obj should be the partition names and their
        corresponding values should contain the loci range and substitution
        model, if any.

        Parameters
        ----------
        dict_obj : OrderedDict
            Ordered dictionary with the definition of the partitions

        Examples
        --------
        Here is an example of a `dict_obj`::

            dict_obj = OrderedDict(("GeneA", [(0,234), "GTR"]),
                                   ("GeneB", [(235, 865), "JC"))
        """

        for k, v in dict_obj:
            # Determining if value contains only the range or the substitution
            # model as well
            if len(v) > 1:
                self.add_partition(k, locus_range=v[0])
            else:
                self.add_partition(k, locus_range=v[0])

    def is_single(self):
        """Returns whether the current `Partitions` has single or multiple
        partitions.

        Returns
        -------
            _ : bool
            Returns True is there is only a single partition defined,
            and False if there are multiple partitions.
        """

        if len(self.partitions) == 1:
            if not [x for x in self.partitions.values()][0][1]:
                return True
            else:
                return False
        else:
            return False

    def _find_parent(self, max_range):
        """Finds the parent partition from a specified range.

        Finds a parent partition of a codon partition.

        Parameters
        ----------
        max_range : int
            The maximum range of the codon partition.

        Returns
        -------
        part : str
            The name of the parent partition, from the `partitions` attribute.
        """

        for part, vals in self.partitions.items():
            lrange = vals[0]
            if lrange[1] == max_range:
                return part

    def add_partition(self, name, length=None, locus_range=None, codon=False,
                      use_counter=False, file_name=None, model_cls=None,
                      auto_correct_name=True):
        """Adds a new partition.

        Adds a new partition providing the length or the range of current
        alignment. If both are provided, the length takes precedence.The range
        of the partition should be in python index, that is, the first position
        should be 0 and not 1.

        Parameters
        ----------
        name : str
            Name of the partition.
        length : int, optional
            Length of the alignment.
        locus_range : list or tuple, optional
            Range of the partition.
        codon : list
            If the codon partitions are already defined, provide the
            starting points in list format, e.g: [1,2,3].
        use_counter : bool
            If True, `locus_range` will be updated according to the `counter`
            attribute.
        file_name : str
            Name of the alignment file.
        model_cls :
            Specified the substitution model that will be set in `models`.
        auto_correct_name : bool
            If set to True, when a partition name already exist, add a counter
            to the end of the name.

        Notes
        -----
        IMPORTANT NOTE on self.model: The self.model attribute was designed
        in a way that allows the storage of different substitution models
        inside the same partition name. This is useful for codon partitions that
        share the same parent partition name. So, for example, a parent
        partition named "PartA" with 3 codon partitions can have a different
        model for each one like this::

            self.models["PartA"] = [[[..model1_params..], [..model2_params..],
                [..model3_params..]], [GTR, GTR, GTR], ["1", "2", "3"]]

        """

        # Check for duplicate names in partitions
        if name in self.partitions:
            if auto_correct_name:
                c = 1
                while "{}_{}".format(name, c) in self.partitions:
                    c += 1
            else:
                raise PartitionException("Partition name %s is already in "
                                         "partition table" % name)

        # When length is provided
        if length:
            # Add to or update alignments_range attribute. This will store the
            # original range of the alignment
            if file_name:
                if file_name in self.alignments_range:
                    current_range = [self.counter, self.counter + (length - 1)]
                    # If start position is earlier than before, update
                    if current_range[0] < self.alignments_range[file_name][0]:
                        self.alignments_range[file_name][0] = current_range[0]
                    # If stop position if later than before, update
                    if current_range[1] > self.alignments_range[file_name][1]:
                        self.alignments_range[file_name][1] = current_range[1]
                else:
                    self.alignments_range[file_name] = [
                        self.counter, self.counter + (length - 1)]

            # Add partition to index list
            self.partitions_index.append([name, 0])
            # Add partition to alignment list
            try:
                self.partitions_alignments[name].append(file_name if file_name
                                                        else name)
            except KeyError:
                self.partitions_alignments[name] = [file_name if file_name else
                                                    name]
            # Create empty model attribute for a single partition
            self.models[name] = [[[]], [None], []]

            self.partitions[name] = [(self.counter,
                                      self.counter + (length - 1)), codon]
            self.counter += length
            self.partition_length += length

        # When a list/tuple range is provided
        elif locus_range:

            if use_counter:
                locus_range = (self.counter,
                               self.counter + locus_range[1] - locus_range[0])

            # Add to or update alignments_range attribute. This will store the
            # original range of the alignment
            if file_name:
                if file_name in self.alignments_range:
                    if locus_range[0] < self.alignments_range[file_name][0]:
                        self.alignments_range[file_name][0] = locus_range[0]
                    if locus_range[1] > self.alignments_range[file_name][1]:
                        self.alignments_range[file_name][1] = locus_range[1]
                else:
                    self.alignments_range[file_name] = list(locus_range)

            # If the maximum range of the current partition is already included
            # in some other partition, and no codon partitions were provided
            # using the "codon" argument, then it should be an undefined codon
            # partition and should be added to an existing partition
            if locus_range[1] <= self.counter and not codon:

                # Find the parent partition
                parent_partition = self._find_parent(locus_range[1])

                if not parent_partition:
                    raise InvalidPartitionFile(
                        "Could not find parent partition of {}. Check the"
                        " ranges of your partitions to ensure no range "
                        "overlaps".format(name))

                # If no codon partition is present in the parent partition,
                # create one
                if not self.partitions[parent_partition][1]:
                    # Add partition to index list
                    self.partitions_index.append([parent_partition, 1])
                    # Create empty model attribute for two partitions
                    self.models[parent_partition] = [[[], []], [None, None], []]

                    parent_start = self.partitions[parent_partition][0][0]
                    self.partitions[parent_partition][1] = [parent_start,
                                                            locus_range[0]]
                else:
                    # Create empty model attribute for additional partitions
                    self.models[parent_partition][0].append([])
                    self.models[parent_partition][1].append(None)

                    # Add partition to index list
                    self.partitions_index.append([parent_partition, 2])

                    self.partitions[parent_partition][1].append(locus_range[0])

            # If the start of the current partition is already within the range
            # of a previous partitions, raise an exception
            elif locus_range[0] < self.counter:
                raise InvalidPartitionFile(
                    "Badly formatted partition with range [{}-{}] starts "
                    "inside the range of a previous partitions ({})".format(
                        locus_range[0], locus_range[1], self.counter))

            # Else, create the new partition. If codon is provided, the codon
            # information is automatically added
            else:
                if model_cls:
                    self.models[name] = model_cls
                else:
                    # Create empty model attribute for a single partition
                    self.models[name] = [[[]], [None], []]
                if codon:
                    self.partitions_index = [[name, x] for x in codon]
                else:
                    # Add partition to index list
                    self.partitions_index.append([name, 0])
                try:
                    self.partitions_alignments[name].append(file_name if
                                                            file_name else name)
                except KeyError:
                    self.partitions_alignments[name] = [file_name if file_name
                                                        else name]

                self.partitions[name] = [(locus_range[0], locus_range[1]),
                                         codon]

                self.counter = locus_range[1] + 1
                self.partition_length = locus_range[1] + 1

    def remove_partition(self, partition_name=None, file_name=None,
                         file_list=None, ns=None):
        """Removes partitions.

        Removes a partitions by a given partition or file name. This will
        handle any necessary changes on the remaining partitions. The changes
        will be straightforward for most attributes, such as partitions_index,
        partitions_alignments and models, but it will require a re-structuring
        of partitions because the ranges of the subsequent partitions will
        have to be adjusted.

        Parameters
        ----------
        partition_name : str
            Name of the partition.
        file_name : str
            Name of the alignment file.
        ns : multiprocesssing.Manager.Namespace
            A Namespace object used to communicate with the main thread
            in TriFusion.
        """

        def rm_part(nm):
            """
            Remove a partition from self.partitions and update the ranges of
            the remaining partitions
            """

            for i in nm:
                del self.partitions[i]

            new_dic = OrderedDict()

            counter = 0
            for p, (nm, vals) in enumerate(self.partitions.items()):

                # Check if the starting position of the next partition is the
                # same as the counter. If so, add the vals to the new dict.
                # Else, correct the ranges based on the counter
                if vals[0][0] == counter:
                    new_dic[nm] = vals
                    counter = vals[0][1] + 1
                else:
                    # Get lenght of the partition
                    part_len = vals[0][1] - vals[0][0]
                    # Create corrected range
                    part_range = (counter, counter + part_len)
                    # Correct codon position start if any
                    if vals[1]:
                        codon = [counter, counter + 1, counter + 2]
                    else:
                        codon = False
                    new_dic[nm] = [part_range, codon]
                    counter = counter + part_len + 1

            return new_dic

        def remove_routine(part_name):
            """
            Routine that removes a partition based on its name. It ca be used
            when calling the remove_partition method with the partition_name
            argument, or with the file_name argument when the partition only
            contains that file name
            """

            if not isinstance(part_name, list):
                part_name = [part_name]

            for p in part_name:
                # Remove partition from partition_index
                self.partitions_index = [
                    x for x in self.partitions_index if x[0] != p]

                # Remove partitions_alignments
                del self.partitions_alignments[p]

                # Remove models
                del self.models[p]

            # Remove from partitions
            self.partitions = rm_part(part_name)

        if partition_name:
            # Raise exception if partition name does not exist
            if partition_name not in self.partitions:
                raise PartitionException("%s is not a partition name" %
                                         partition_name)

            remove_routine(partition_name)

        if file_list:

            part_list = []
            for part, fl in self.partitions_alignments.items():
                if any((x for x in fl if x in file_list)):
                    part_list.append(part)

            # Not this, will remove a partition containing multiple alignments
            # if at least one is in the file_list argument
            remove_routine(part_list)

        if file_name:

            # Set file_found to True, when there is a match. If no match is
            # found, raise a PartitionException at the end of the loop.
            file_found = False
            for part, file_list in self.partitions_alignments.items():
                if file_name in file_list:
                    file_found = True
                    # If the partitions consists only of the provided file,
                    # Remove the entire partition
                    if len(file_list) == 1:
                        remove_routine(part)
                    # If the partition contains other files, then only remove
                    # the current file from the partition
                    else:
                        self.partitions_alignments[part].remove(file_name)

            if not file_found:
                raise PartitionException("%s file does not belong to any"
                                         "partition" % file_name)

    def change_name(self, old_name, new_name):
        """Changes name of a partition.

        Parameters
        ----------
        old_name : str
            Original partition name.
        new_name : str
            New partition name.
        """

        self.partitions[new_name] = self.partitions.pop(old_name)
        self.partitions_alignments[new_name] = \
            self.partitions_alignments.pop(old_name)
        self.models[new_name] = self.models.pop(old_name)

    def merge_partitions(self, partition_list, name):
        """Merges multiple partitions into a single one.

        Parameters
        ----------
        partition_list : list
            List with partition names to be merged.
        name : str
            Name of new partition
        """

        def merger(ranges):
            """
            Generator that merges ranges in a list of tuples. For example,
            if ranges is [(1, 234), (235, 456), (560, 607), (607,789)]
            this generator will yield [(1, 456), (560, 789)]
            """
            previous = 0
            last_start = 0
            for st, en in ranges:
                if not previous:
                    last_start = st
                    previous = en
                elif st - 1 == previous:
                    previous = en
                else:
                    yield last_start, previous
                    previous = en
                    last_start = st

            yield last_start, en

        def flatter(s):
            """
            Creates a flat iterator of tuples. If s is [[(1,2), (2,3)], (4,5)]
            this will yield ((1,2), (2,3), (4,5))
            """
            for i in s:
                if isinstance(i, tuple):
                    yield i
                else:
                    for j in i:
                        yield j

        # Get new range
        new_range = [x for x in merger(flatter((y[0] for x, y in
                                               self.partitions.items()
                                               if x in partition_list)))]

        # Add entries for new partition
        self.partitions[name] = [new_range[0] if len(new_range) == 1 else
            new_range, False]
        self.partitions_alignments[name] = list(set([i for x, y in
                                            self.partitions_alignments.items()
                                            if x in partition_list for i in y]))
        self.models[name] = [[[]], [None], []]

        # Delete previous partitions and update merged dict
        for p in partition_list:
            if len(self.partitions_alignments[p]) == 1:
                self.merged_files[self.partitions_alignments[p][0]] = \
                    self.partitions[p][0]
            del self.partitions[p]
            del self.partitions_alignments[p]
            del self.models[p]

    def split_partition(self, name, new_range=None, new_names=None):
        """Splits one partition into two.

        Splits a partitions with `name` into two with the tuple list provided
        by `new_range`. If new_range is None, This will split the partition
        by its alignment files instead.

        Parameters
        ----------
        name : str
            Name of the partition to be split.
        new_range : list or tuple, optional
            List of two tuples, containing the ranges of the new partitions.
        new_names : list, optional
            The names of the new partitions.
        """

        if new_range:

            # Add new partitions
            for n, r in zip(new_names, new_range):
                self.partitions[n] = [r, False]
                # Create new partitions_alignments. Keep the original alignment
                # file for both
                self.partitions_alignments[n] = self.partitions_alignments[name]
                self.models[n] = [[[]], [None], []]

        else:

            for aln in self.partitions_alignments[name]:
                #  Get original range of alignment file
                new_range = self.merged_files[aln]
                # Add new partitions
                aln_name = basename(aln)
                self.partitions[aln_name] = [new_range, False]
                self.partitions_alignments[aln_name] = [aln]
                self.models[aln_name] = [[[]], [None], []]

        # Delete original partition
        del self.partitions[name]
        del self.partitions_alignments[name]
        del self.models[name]

    # ==========================================================================
    # Model handling
    # ==========================================================================

    def parse_nexus_model(self, string):
        """Parses a substitution model defined in a prset and/or lset command.

        Parameters
        ----------
        string : str
            String with the prset or lset command.
        """

        string = string.lower()

        # Find out which partitions the current parameters apply to. If
        # detected, it should be something like "applyto=(1,2)"
        applyto = re.findall(r"applyto=\(.*\)", string)
        # Find parameters
        nst = re.findall(r"nst=[0-9]", string)
        statefreqpr = re.findall(r"statefreqpr=.*\)", string)

        # Collect params
        params = [x[0] for x in [nst, statefreqpr] if x]

        if applyto:
            if applyto == ["applyto=(all)"]:
                for partition in self.partitions:
                    self.models[partition][0] += params
            else:
                # Get target partitions
                part_index = [int(x) for x in
                              re.split("[()]", applyto[0])[1].split(",")]
                for i in part_index:
                    part = self.partitions_index[i - 1]
                    # Get partition name
                    part_name = part[0]
                    # Get subpartition index. 0 if single partition, other if
                    # multiple subpartition
                    part_subpart = part[1]
                    self.models[part_name][0][part_subpart] += params

    def get_model_name(self, params):
        """Given a list of parameters, return the name of the model

        Parameters
        ----------
        p : list
            List of prset/lset parameters

        Returns
        -------
        model : str or None
            Returns the name of the model if it finds. Else, returns None.
        """

        for model, p in self._models["mrbayes"].items():
            if params == p:
                return model
        else:
            return None

    def set_model(self, partition, models, links=None, apply_all=False):
        """Sets substitution model for a given partition.

        Parameters
        ----------
        partition : str
            Partition name.
        models : list
            Model names for each of the three codon partitions. If there
            are no codon partitions, provide only a single element to the list.
        links : list
            Provide potential links between codon models. For
            example, if codon 1 and 2 are to be linked, it should be:
            links=["12", "3"]
        apply_all : bool
            If True, the current model will be applied to all partitions.
        """

        # Get list with partitions to be changed
        if apply_all:
            plist = [x for x in self.partitions]
        else:
            plist = [partition]

        # Replace "No model" string with None
        models = [None if x == "No model" else x for x in models]

        # Set model to the whole partition
        if len(models) == 1:
            # If the current partition was previously defined as having codon
            # partitions, revert it
            for p in plist:
                if self.partitions[p][1]:
                    self.partitions[p][1] = False
                self.models[p][1] = models

        # Set codon models
        else:
            for p in plist:
                # Change the partition in self.partitions to have codon
                # partitions
                st_idx = self.partitions[p][0][0]
                self.partitions[p][1] = [st_idx + x for x in range(3)]
                self.models[p][1] = models
                self.models[p][2] = links

    def write_to_file(self, output_format, output_file, model="LG"):
        """Writes partitions to a file.

        Writes the Partitions object into an output file according to the
        output_format. The supported output formats are RAxML and Nexus. The
        `model` option is for the RAxML format only.

        Parameters
        ----------
        output_format : str
            Output format of partitions file. Can be either "nexus" or
            "raxml".
        output_file : str
            Path to output file.
        model : str
            Name of the model for the partitions. "raxml" format only.
        """

        if output_format == "raxml":
            outfile_handle = open(output_file + ".part.File", "w")
            for part, rge in self.partitions.items():
                partition_range = "-".join([str(x + 1) for x in rge[0]])
                outfile_handle.write("%s, %s = %s\n" % (model,
                                                        part,
                                                        partition_range))

            outfile_handle.close()

        elif output_format == "nexus":
            outfile_handle = open(output_file + ".charset", "w")
            for part, rge in self.partitions.items():
                outfile_handle.write("charset %s = %s;\n" % (
                                     part,
                                     "-".join([str(x + 1) for x in rge[0]])))

            outfile_handle.close()

        return 0


class Zorro(object):
    """
    Class that handles the concatenation of zorro weights.

    Parameters
    ----------
    alignment_list : trifusion.process.sequence.AlignmentList
        AlignmentList object.
    suffix : str
        Suffix of the zorro weight files, based on the corresponding
        input alignments.
    zorro_dir : str
        Path to directory where zorro weight files are stored.
    """

    def __init__(self, alignment_list, suffix="_zorro.out", zorro_dir=None):

        self.weigth_values = []
        self.suffix = suffix

        for file_path in [x.path for x in alignment_list.alignments.values()]:
            # If zorro_dir is provided, use the specified path
            if zorro_dir:
                zorro_file = splitext(file_path.split(sep)[-1])[0]
                zorro_file = "{}{}.txt".format(join(zorro_dir, zorro_file), suffix)
            # If zorro_dir is not provided, use the same path as the input alignment
            else:
                zorro_file = file_path.split(".")[0] + self.suffix + ".txt"
            # alignment file is shared with the corresponding zorro file
            zorro_handle = open(zorro_file)
            self.weigth_values += [int(round(float(weigth.strip()))) for
                                   weigth in zorro_handle]

    def write_to_file(self, output_file):
        """ Creates a concatenated file with the zorro weights for the
        corresponding alignment files."""
        outfile = output_file + "_zorro.out"
        outfile_handle = open(outfile, "w")
        for weigth in self.weigth_values:
            outfile_handle.write("%s\n" % weigth)
        outfile_handle.close()


__author__ = "Diogo N. Silva"



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
#
#  Copyright 2012 Unknown <diogo@arch>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

from argparse import ArgumentTypeError
import os

try:
    from process.base import print_col, RED, YELLOW
except ImportError:
    from trifusion.process.base import print_col, RED, YELLOW


def triseq_arg_check(arg):

    if arg.gcoder and "nexus" not in arg.output_format:
        print_col("Gap coding can only be performed for Nexus output format.",
                  RED)

    if arg.gcoder and "nexus" in arg.output_format and \
            arg.output_format != ["nexus"]:
        print_col("Gap coding can only be performed for Nexus output format."
                  " This operation will be ignored for other output formats.",
                  YELLOW, quiet=arg.quiet)

    if arg.conversion and arg.reverse:
        print_col("Ignoring conversion flag (-c) when specifying reverse"
                  " concatenation (-r)", YELLOW, quiet=arg.quiet)

    if arg.outfile and arg.reverse:
        print_col("Ignoring output file option (-o) when specifying reverse"
                  " concatenation (-r)", YELLOW, quiet=arg.quiet)

    if arg.partition_file is not None and arg.outfile is None:
        print_col("An output file must be provided with option '-o'", RED)

    if "ima2" in arg.output_format and arg.ima2_params is None:
        print_col("Additional arguments must be provided with the"
                  " option --ima2-params when selecting ima2 output format",
                  RED)

    if "ima2" in arg.output_format and len(arg.ima2_params) != 4:
        print_col("Four additional arguments must be provided with"
                  " option --ima2-params when selecting the "
                  "ima2 output format. %s were given" %
                  (len(arg.ima2_params)), RED)

    if arg.partition_file is not None:
        return 0

    if arg.conversion is None and arg.outfile is None and arg.reverse is None\
            and arg.select is None and arg.get_taxa is False:

        print_col(
            "If you wish to concatenate, provide the output file name using "
            "the '-o' option. If you wish to convert a "
            "file, specify it using the '-c' option", RED)

    if len(arg.infile) == 1 and arg.conversion is None and arg.reverse is None\
            and arg.collapse is None:

        print_col(
            "Cannot perform concatenation of a single file. Please provide"
            " additional files to concatenate, or specify the conversion "
            "'-c' option", RED)

    if arg.zorro is not None and len(arg.infile) == 1:
        print_col(
            "The '-z' option cannot be invoked when only a single input "
            "file is provided. This option is reserved for"
            " concatenation of multiple alignment files", RED)

    if arg.consensus and arg.output_format != ["fasta"]:
        print_col("Output format must be only Fasta when using the "
                  "consensus option", RED)

    if not arg.consensus and arg.consensus_single:
        print_col("Ignoring consensus single file option (--consensus-single-"
                  "file) when the consensus operation is not specified",
                  YELLOW, quiet=arg.quiet)

    else:
        return 0


def post_aln_checks(arg, aln_obj):

    if arg.consensus == ["IUPAC"] and aln_obj.sequence_code[0] != "DNA":
        print_col("'IUPAC' option of the consensus operation can "
                  "only be performed on nucleotide alignments.", RED)
    if arg.codon_filter and aln_obj.sequence_code[0] != "DNA":
        print_col("The codon filter option (--codon-filter) can only be"
                  " performed on nucleotide alignments.", RED)
    if aln_obj.bad_alignments:
        print_col("The following input files could not be read or are empty"
                  ": {}".format(" ".join(aln_obj.bad_alignments)), YELLOW)
    if aln_obj.non_alignments:
        print_col("The following input files have alignments of unequal "
                  "length: {}".format(" ".join(aln_obj.non_alignments)),
                  YELLOW)

    else:
        return 0


def mfilters(filt):
    """
    Checks the type of the some filter options. Assures that the values
    are within the accepted boundaries of [0-100]. If the provided filter
    is a decimal, it converts to percentage automatically
    :param filt: (string) The value provided with the --missing-filter
    :return:
    """

    # Check if filt is convertable to float
    try:
        filt = float(filt)
    except ValueError:
        raise ArgumentTypeError("The value '{}' is not a "
                                "number between 0 and 100.".format(filt))

    # Check if filt is within acceptable range
    if not (0 <= filt <= 100):
        raise ArgumentTypeError("The value '{}' must be a number between "
                                "0 and 100.".format(int(filt)))

    # If filt is a float between 0 and 1, convert to percentage
    if 0 <= filt < 1:
        filt = int(filt * 100)
    else:
        filt = int(filt)

    return filt


def check_infile_list(infiles):

    dirs = []
    lost = []
    good_files = []

    for fpath in infiles:

        if not os.path.exists(fpath):
            lost.append(fpath)

        elif os.path.isdir(fpath):
            dirs.append(fpath)

        else:
            good_files.append(fpath)

    return good_files, dirs, lost



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
#
#  Copyright 2012 Unknown <diogo@arch>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import warnings

# Suppress import warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore")

    import os
    import sys
    import shutil
    import time
    import argparse
    from glob import glob

    try:
        from process.base import print_col, RED, GREEN, YELLOW, CleanUp
        from process import sequence as seqset
        from process import data
        from process.error_handling import *
        from base.sanity import triseq_arg_check, mfilters, post_aln_checks, \
            check_infile_list
        from progressbar import ProgressBar, Timer, Bar, Percentage, \
            SimpleProgress
    except ImportError:
        from trifusion.process.base import print_col, RED, GREEN, YELLOW,\
            CleanUp
        from trifusion.process import sequence as seqset
        from trifusion.process import data
        from trifusion.process.error_handling import *
        from trifusion.base.sanity import triseq_arg_check, mfilters, \
            post_aln_checks, check_infile_list
        from trifusion.progressbar import ProgressBar, Timer, Bar,\
            Percentage, SimpleProgress


def gen_wgt(msg):

    bar_wdg = [
        "( ", SimpleProgress(), " ) ",
        Bar(),
        Percentage(),
        " [", Timer(), "] ",
    ]

    return bar_wdg

@CleanUp
def main_parser(arg, alignment_list):
    """ Function with the main operations of TriSeq """

    print_col("Executing TriSeq module at %s %s" % (
        time.strftime("%d/%m/%Y"), time.strftime("%I:%M:%S")), GREEN,
              quiet=arg.quiet)

    # Create temp directory
    tmp_dir = ".trifusion-temp"
    if not os.path.exists(tmp_dir):
        os.makedirs(tmp_dir)

    # Set path to temporary sqlite database
    sql_db = os.path.join(tmp_dir, "trifusion.db")

    # If database already exists, erase it. Make sure we start fresh.
    if os.path.exists(sql_db):
        os.remove(sql_db)

    # Defining main variables
    conversion = arg.conversion
    output_format = arg.output_format
    outfile = arg.outfile
    interleave = arg.interleave
    model_phy = arg.model_phy
    # outgroup_taxa = arg.outgroup_taxa

    # Defining output file name
    if conversion is None and arg.outfile is not None:
        outfile = "".join(arg.outfile)
    elif conversion is None and arg.outfile is not None:
        outfile = "".join(arg.outfile)
    elif arg.consensus and arg.consensus_single and not arg.outfile:
        outfile = "consensus"

    # The input file at this stage is not necessary
    # If just converting the partition file format do this and exit
    if arg.partition_file is not None:
        # Initializing Partitions instance and reading partitions file
        partition = data.Partitions()
        partition.read_from_file(arg.partition_file)
        if partition.partition_format == "nexus":
            partition.write_to_file("raxml", outfile, model_phy)
        else:
            partition.write_to_file("nexus", outfile)
        return 0

    # Support wildcars as arguments for windows
    fl = []
    if sys.platform in ["win32", "cygwin"]:
        for p in alignment_list:
            fl += glob(p)
        alignment_list = fl

    # Check input files for directories
    alignment_list, dirs, lost = check_infile_list(alignment_list)

    if dirs:
        print_col("Ignoring input files pointing to a directory: {}".format(
            " ".join(dirs)), YELLOW)
    if lost:
        print_col("Ignoring input files that do not exist: {}".format(
            " ".join(lost)), YELLOW)
    if not alignment_list:
        print_col("No valid input files have been provided. Terminating...",
                  RED)

    # Input alignments are mandatory from now on
    if not arg.quiet:
        pbar = ProgressBar(max_value=len(alignment_list), widgets=gen_wgt(""))
    else:
        pbar = None

    print_col("Parsing %s alignments" % len(alignment_list), GREEN,
              quiet=arg.quiet)
    alignments = seqset.AlignmentList(alignment_list, sql_db=sql_db,
                                      pbar=pbar)

    post_aln_checks(arg, alignments)

    # ################################ Utilities ##############################
    # Return a file with taxa list and exit
    if arg.get_taxa is True:
        print_col("Writing taxa to new file", GREEN, quiet=arg.quiet)
        alignments.write_taxa_to_file()
        return 0

    # Remove taxa
    if arg.remove:
        print_col("Removing taxa", GREEN, quiet=arg.quiet)
        alignments.remove_taxa(arg.remove)

    # Grep taxa
    if arg.grep:
        print_col("Grepping taxa", GREEN, quiet=arg.quiet)
        alignments.remove_taxa(arg.grep, mode="inverse")

    # Select alignments
    if arg.select:
        print_col("Selecting alignments", GREEN, quiet=arg.quiet)
        if not os.path.exists("Taxa_selection"):
            os.makedirs("Taxa_selection")

        # Check if any of the provided taxa is absent from the alignments
        absent_taxa = [x for x in arg.select if x not in alignments.taxa_names]
        if absent_taxa:
            print_col("The following taxa were not found in any alignment and"
                      " will be ignored: {}".format(" ".join(absent_taxa)),
                      YELLOW, quiet=arg.quiet)

        selected_alignments = alignments.select_by_taxa(arg.select,
                                                        mode="relaxed")
        for aln in selected_alignments:
            alignment_file = aln.path
            shutil.copy(alignment_file, "Taxa_selection")

        return

    # ############################# Main operations ###########################
    # Reverse concatenation
    if arg.reverse is not None:
        print_col("Reverse concatenating", GREEN, quiet=arg.quiet)
        if len(alignment_list) > 1:
            raise ArgumentError("Only one input file allowed for reverse "
                                "concatenation")
        if arg.reverse:
            aln = alignments.alignments.values()[0]
            # Initializing and reading partition file
            partition = data.Partitions()
            partition.read_from_file(arg.reverse[0])
            # Updating alignment _partitions
            aln.set_partitions(partition)
            alignments.set_partition_from_alignment(aln, reset=True)

        alignments.reverse_concatenate(pbar=pbar)

    # Filtering
    # Filter by minimum taxa
    if arg.min_taxa:
        print_col("Filtering by minimum taxa", GREEN, quiet=arg.quiet)
        alignments.filter_min_taxa(arg.min_taxa, pbar=pbar)

    # Filter by alignments that contain taxa
    if arg.contain_filter:
        print_col("Filtering alignment(s) including a taxa group", GREEN,
                  quiet=arg.quiet)
        alignments.filter_by_taxa(arg.contain_filter, "Contain", pbar=pbar)

    # Filter by alignments that exclude taxa
    if arg.exclude_filter:
        print_col("Filtering alignments excluding a taxa group", GREEN,
                  quiet=arg.quiet)
        alignments.filter_by_taxa(arg.exclude_filter, "Exclude", pbar=pbar)

    # Filter by codon position
    if arg.codon_filter:
        print_col("Filtering by codon positions", GREEN, quiet=arg.quiet)
        if alignments.sequence_code[0] == "DNA":
            codon_settings = [True if str(x) in arg.codon_filter else False
                              for x in range(1, 4)]
            alignments.filter_codon_positions(codon_settings, pbar=pbar)

    # Filter by missing data
    if arg.m_filter:
        print_col("Filtering by missing data", GREEN, quiet=arg.quiet)
        alignments.filter_missing_data(arg.m_filter[0], arg.m_filter[1],
                                       pbar=pbar, use_main_table=True)

    # Filtering by variable sites
    if arg.var_filter:
        print_col("Filtering by variable sites", GREEN, quiet=arg.quiet)
        alignments.filter_segregating_sites(arg.var_filter[0],
                                            arg.var_filter[1],
                                            pbar=pbar)

    # Filtering by informative sites
    if arg.inf_filter:
        print_col("Filtering by variable sites", GREEN, quiet=arg.quiet)
        alignments.filter_informative_sites(arg.inf_filter[0],
                                            arg.inf_filter[1],
                                            pbar=pbar)

    # Concatenation
    if not arg.conversion and arg.reverse is None and not arg.consensus:
        print_col("Concatenating", GREEN, quiet=arg.quiet)
        alignments.concatenate(pbar=pbar)

        # Concatenate zorro files
        if arg.zorro:
            zorro = data.Zorro(alignment_list, arg.zorro)
            zorro.write_to_file(outfile)

    # Collapsing
    if arg.collapse:
        print_col("Collapsing", GREEN, quiet=arg.quiet)
        alignments.collapse(use_main_table=True, pbar=pbar,
                            haplotypes_file=outfile)

    # Gcoder
    if arg.gcoder:
        print_col("Coding gaps", GREEN, quiet=arg.quiet)
        if output_format == ["nexus"]:
            alignments.code_gaps(use_main_table=True, pbar=pbar)

    # Consensus
    if arg.consensus:
        consensus_type = arg.consensus[0]
        print_col("Creating consensus sequences", GREEN, quiet=arg.quiet)
        alignments.consensus(consensus_type, single_file=arg.consensus_single,
                             pbar=pbar)

    # Write output
    print_col("Writing output", GREEN, quiet=arg.quiet)
    alignments.write_to_file(output_format, output_file=outfile,
                             interleave=interleave,
                             ima2_params=arg.ima2_params,
                             partition_file=True,
                             use_charset=True,
                             pbar=pbar)


def get_args(arg_list=None, unittest=False):

    # The inclusion of the argument definition in main, makes it possible to
    # import this file as a module and not triggering argparse. The
    # alternative of using a if __name__ == "__main__" statement does not
    # work well with the entry_points parameter of setup.py, since they call
    # the main function but do nothing inside said statement.
    parser = argparse.ArgumentParser(description="Command line interface for "
                                                 "TriFusion Process module")

    # Main execution
    main_exec = parser.add_argument_group("Main execution")
    main_exec.add_argument("-in", dest="infile", nargs="+", help="Provide the "
                           "input file name. If multiple files are provided"
                           ", please separated the names with spaces")
    main_exec.add_argument("-of", dest="output_format", nargs="+",
                           default=["nexus"],
                           choices=["nexus", "phylip", "fasta", "mcmctree",
                                    "ima2", "stockholm", "gphocs"],
                           help="Format of the output file(s). You may select "
                           "multiple output formats simultaneously (default is"
                           " '%(default)s')")
    main_exec.add_argument("-o", dest="outfile", help="Name of the output file")

    # Alternative modes
    alternative = parser.add_argument_group("Alternative execution modes")
    alternative.add_argument("-c", dest="conversion", action="store_const",
                             const=True, help="Used for conversion of the "
                             "input files passed as arguments with the -in "
                             "option. This flag precludes the usage of the -o "
                             "option, as the output file name is automatically "
                             "generated based on the input file name")
    alternative.add_argument("-r", dest="reverse", help="Reverse a concatenated"
                             " file into its original single locus alignments."
                             " A partition file similar to the one read by "
                             "RAxML must be provided", nargs="*")
    alternative.add_argument("-z", "--zorro-suffix", dest="zorro", type=str,
                             help="Use this option if you wish to concatenate "
                             "auxiliary Zorro files associated with each "
                             "alignment. Provide the sufix for the concatenated"
                             " zorro file")
    alternative.add_argument("-p", "--partition-file", dest="partition_file",
                             type=str, help="Using this option and providing "
                             "the partition file will convert it between a "
                             "RAxML or Nexus format")
    alternative.add_argument("-s", dest="select", nargs="*", help="Selects "
                             "alignments containing the provided taxa "
                             "(separate multiple taxa with whitespace)")
    alternative.add_argument("--collapse", dest="collapse",
                             action="store_const",
                             const=True, default=False, help="Use this flag if "
                             "you would like to collapse the input alignment(s)"
                             " into unique haplotypes")
    alternative.add_argument("--code-gaps", dest="gcoder", action="store_const",
                             const=True, default=False, help="Use this flag to "
                             "code the gaps of the alignment into a binary "
                             "state  matrix that is appended to the end of "
                             "the alignment")
    alternative.add_argument("--consensus", dest="consensus", nargs=1,
                             choices=["First sequence", "IUPAC", "Soft mask",
                                      "Remove"],
                             help="Creates a consensus of the final alignments"
                             " specifying how variation is handled")
    alternative.add_argument("--consensus-single-file",
                             dest="consensus_single", action="store_const",
                             const=True, default=False, help="Merges "
                             "consensus sequences in a single file")

    filter_g = parser.add_argument_group("Filter options")
    filter_g.add_argument("--missing-filter", dest="m_filter", nargs=2,
                          help="Use this  option if you wish to filter the"
                          " alignment's missing data. Along with this "
                          "option provide the threshold percentages for "
                          "gap and missing data, respectively (e.g. -filter "
                          "50 75 - filters alignments columns with more "
                          "than 50%% of gap+missing data and columns with "
                          "more than 75%% of true missing data)",
                          type=mfilters)
    filter_g.add_argument("--min-taxa", dest="min_taxa",
                          type=mfilters, help="Set the minimum percentage "
                          "of taxa that needs to be present in an alignment")
    filter_g.add_argument("--contain-taxa", dest="contain_filter",
                          nargs="*", help="Only "
                          "processes alignments that contain the specified "
                          "taxa")
    filter_g.add_argument("--exclude-taxa", dest="exclude_filter", help="Only "
                          "process alignments that do NOT contain the "
                          "specified taxa")
    filter_g.add_argument("--codon-filter", dest="codon_filter", nargs="*",
                          choices=["1", "2", "3"], help="Include only the "
                          "codon positions specified by this option (DNA only)")
    filter_g.add_argument("--variable-filter", dest="var_filter", nargs=2,
                          help="Provide minimum and maximum values of "
                          "variable sites for each alignment. Filters "
                          "alignments with a number of variable sites outside "
                          "the specified range", type=int)
    filter_g.add_argument("--informative-filter", dest="inf_filter", nargs=2,
                          help="Provide minimum and maximum values of "
                          "informative sites for each alignment. Filters "
                          "alignments with a number of informative sites "
                          "outside the specified range", type=int)

    # Formatting options
    formatting = parser.add_argument_group("Formatting options")
    formatting.add_argument("--model", dest="model_phy", default="LG",
                            choices=["DAYHOFF", "DCMUT", "JTT", "MTREV", "WAG",
                                     "RTREV", "CPREV", "VT", "BLOSUM62",
                                     "MTMAM", "LG"],
                            help="This option only applies for the "
                            "concatenation  of protein data into phylip "
                            "format. Specify the model for all partitions "
                            "defined in the partition file  (default is '%("
                            "default)s')")
    formatting.add_argument("--interleave", dest="interleave",
                            action="store_const", const="interleave",
                            default=False, help="Specify this  option to "
                            "write output files in interleave format (currently"
                            " only supported for nexus files")
    formatting.add_argument("--ima2-params", dest="ima2_params", nargs=4,
                            help="Provide 4 additional arguments needed to "
                            "write the output in a format compliant with "
                            "IMa2. The order of the required arguments "
                            "(separated by whitespace is as follows: "
                            "[(1) File name of population mapping]"
                            "[(2) Population tree]"
                            "[(3) Mutational model]"
                            "[(4) Inheritance Scalar]. "
                            "Additional notes: (1) The  population mapping "
                            "file is a simple .csv file  containing two "
                            "columns separated by a semi-colon, "
                            "in which the first column contains the taxon "
                            "name and the second column contains the "
                            "corresponding population name; (2) The order of "
                            "the population names in the population tree "
                            "must be the same as the order in the file with "
                            "the population mapping")

    # Data manipulation
    manipulation = parser.add_argument_group("Data manipulation")
    manipulation.add_argument("-rm", dest="remove", nargs="*", help="Removes "
                              "the  specified taxa from the final alignment. "
                              "Unwanted taxa my be provided in a csv file "
                              "containing 1 column with a species name in "
                              "each line or they may be specified in the "
                              "command line and separated by whitespace")
    manipulation.add_argument("-grep", dest="grep", nargs="*", help="The "
                              "inverse  of the -rm command. It removes all "
                              "taxa from the alignment except for the ones "
                              "specified with this option. Taxa names may be "
                              "specified in a csv file containing 1 column "
                              "with a species name in each line or in the "
                              "command line separated by whitespace")
    # manipulation.add_argument("-outgroup", dest="outgroup_taxa", nargs="*",
    #                           help="Provide taxon names/number for the "
    #                           "outgroup  (This option is only supported for "
    #                           "NEXUS output format files)")

    utilities = parser.add_argument_group("Utilities")
    utilities.add_argument("--get-taxa", dest="get_taxa",
                           action="store_const", const=True,
                           default=False, help="Writes all taxa names into a"
                           " .csv file")

    miscellaneous = parser.add_argument_group("Miscellaneous")
    miscellaneous.add_argument("-quiet", dest="quiet", action="store_const",
                               const=True, default=False, help="Removes all "
                               "terminal output")

    args = parser.parse_args(arg_list)

    # Print help when no arguments are provided
    if len(sys.argv) == 1 and not unittest:
        parser.print_help()
        sys.exit(1)

    return args


def main():
    arguments = get_args()
    triseq_arg_check(arguments)
    main_parser(arguments, arguments.infile)


if __name__ == "__main__":

    main()


__author__ = "Diogo N. Silva"



"""
This file contains the color scheme for some widgets and backgrounds of
TriFusion
"""

# Main colors for general usage in the application. They redirect to _blue and
# _red in app.py
c_app_blue = (0.216, 0.67, 0.784, 1)
c_app_red = (1., 0.39, 0.39, 1.)

# Main window background
c_main_background = (.08, .08, .08, 1)

# Sidepanel background
c_iner_side_panel_background = (0.1, 0.1, 0.1, .9)
c_outer_side_panel_background = c_iner_side_panel_background

# Sidepanel Caroussel headers
c_side_panel_headers = (.3, .3, .3, 1)

# Light boxes
c_light_box = (.2, .2, .2, 1)

# Dark boxes
c_dark_box = (0, 0, 0, 0)

# Orthology sidepanel
c_orto_sidepanel = (.2, .2, .2, .5)

# Statistics side panel
c_stats_sidepanel = (0.1, 0.1, 0.1, .7)

# Popup background color
c_popup_background = c_main_background
c_popup_border = (0.216, 0.67, 0.784, 1)

# Filechoosers headers
c_filechooser_headers = (.2, .2, .2, .9)

# Dropdown backgrounds
c_dropdown_background = (.15, .15, .15, 1)



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
#
#  Copyright 2012 Unknown <diogo@arch>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import matplotlib
# Using the agg backend prevent a segmentation fault when quiting the
# application
matplotlib.use("agg")

import matplotlib.pyplot as plt
import matplotlib.cm as cm
from scipy.stats import spearmanr
from scipy.interpolate import spline
import numpy as np
from itertools import chain
from collections import Counter
import seaborn as sns
import os
import functools

# Set of 10 easily distinguishable colors that will be used when generating
# plots to ensure consistency in color usage. In case more than 10 colors
#  are required (ugh...) they will be randomly generated henceforth.

clr_list = [[0, .53, .66],  # light blue
            [1, .16, .16],     # light red
            [.21, .78, .22],  # light green
            [.83, .55, .37],  # light brown
            [.55, .37, .83],  # purple
            [1, .4, 0],     # light orange
            [.77, .77, .77],  # light grey
            [.78, .22, .22],  # dark red
            [.1, .1, .1],     # dark grey
            [0, .66, 0]]      # green

two_clr_list = [[0, .53, .66],
                [.62, .80, .85]]


def set_props(func):
    """Decorator used to set general plot attributes

    This decorator is use to automatically set several plot properties based
    on the arguments provided to the plotting functions. In this way,
    there is no need to write repetitive code in each function. The accepted
    arguments are:

      - ax_names : list.
        Axis names. First element is x-axis label, second is y-axis label
      - title : str.
        Title of the plot
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):

        plt.clf()
        plt.close()

        res = func(*args, **kwargs)

        # Set axis names
        override = {
            "fontsize": 14,
            "fontweight": "bold",
            "color": "grey"
        }
        if "ax_names" in kwargs:
            if kwargs["ax_names"][0]:
                plt.xlabel(kwargs["ax_names"][0], labelpad=15, **override)
            if kwargs["ax_names"][1]:
                plt.ylabel(kwargs["ax_names"][1], labelpad=15, **override)

        # Set title
        if "title" in kwargs:
            plt.title(kwargs["title"], y=1.05, **override)

        return res

    return wrapper


@set_props
def scatter_plot(data, correlation=False, ax_names=None, table_header=None,
                 title=None):
    """Generates a scatter plot from a 2D array.

    Builds a scatter plot from a 2D array. Also calculates the
    correlation coefficient if requested by the `correlation` argument.

    Parameters
    ----------
    data : numpy.array
        2D array containing the x and y data points.
    correlation : bool
        If True, the spearman's rank correlation coefficient is calculated
        and added to the plot as an annotation
    ax_names : list
        List with the labels for the x-axis (first element) and y-axis
        (second element).
    table_header : list
        List with the header of the table object. Each element represents
        a column.
    title : str
        Title of the plot.

    Returns
    -------
    fig : matplotlib.pyplot.Figure
        Figure object of the plot.
    _ : None
        Placeholder for the legend object. Not used here but assures
        consistency across other plotting methods.
    table : list
        Table data in list format. Each item in the list corresponds to a
        table row.
    """

    plt.rcParams["figure.figsize"] = (8, 6)

    # Use ggpot style
    plt.style.use("ggplot")

    fig, ax = plt.subplots()

    # Generate scatter plot
    ax.scatter(data[0], data[1], alpha=0.5, edgecolors=None)

    # Plot best fit line
    ax.plot(data[0], np.poly1d(np.polyfit(data[0], data[1], 1))(data[0]))

    # Set plot limits
    max1 = max(data[0])
    max2 = max(data[1])
    ax.set_xlim(0, max1 + (max1 * .1))
    ax.set_ylim(0, max2 + (max2 * .1))

    # Calculate spearman's rank correlation coefficient and add it to the plot
    if correlation:
        rho, pval = spearmanr(data[0], data[1])

        ax.text((ax.get_xlim()[1] - ax.get_xlim()[0]) * .05,
                ax.get_ylim()[1] - (ax.get_ylim()[1] * .05),
                r"$\rho$ = {}; p-value = {}".format(round(rho, 4),
                                                    round(pval, 4)))

    # Generate table data
    if table_header:
        table = [table_header]
    else:
        table = []

    for i, j in zip(*data):
        table.append([i, j])

    return fig, None, table


@set_props
def bar_plot(data, labels=None, title=None, ax_names=None,
             lgd_list=None, reverse_x=False, table_header=None):
    """Creates a bar plot from a `data` array.

    If a multi-dimensional array is provided, it will create a stacked bar
    plot.

    Parameters
    ----------
    data : numpy.array
        Single or multi-dimensional array with plot data.
    labels : list
        List of xtick labels. Should have the same length as `data`.
    title : str
        Title of the plot.
    ax_names : list
        List with the labels for the x-axis (first element) and y-axis
        (second element).
    lgd_list : list
        For categorical plots, provide the label of each category.
    reverse_x : bool
        If True, reverse the x-axis orientation.
    table_header : list
        List with the header of the table object. Each element represents
        a column.

    Returns
    -------
    fig : matplotlib.Figure
        Figure object of the plot.
    lgd : matplotlib.Legend
        Legend object of the plot.
    table : list
        Table data in list format. Each item in the list corresponds to a
        table row.
    """

    if len(labels) > 10:
        plt.rcParams["figure.figsize"] = (len(labels) / 3, 6)
    else:
        plt.rcParams["figure.figsize"] = (8, 6)

    # Use ggpot style
    plt.style.use("ggplot")

    fig, ax = plt.subplots()

    # Create plot
    for i, d in enumerate(data):
        # Get color from 10 color list. If more than 10 colors are required,
        # randomly generate new ones
        if len(data) > 2:
            try:
                clr = clr_list[i]
            except IndexError:
                clr = np.random.rand(3, 1)
        else:
            clr = two_clr_list[i]

        # Create plot for first entry
        if d == data[0]:
            ax.bar(np.arange(len(d)), d, align="center", color=clr,
                   label=lgd_list[i] if lgd_list else None)
        # Add plots on top
        else:
            ax.bar(np.arange(len(d)), d, align="center", color=clr,
                   bottom=data[i - 1],
                   label=lgd_list[i] if lgd_list else None)

    # Set legend
    if lgd_list:
        lgd = plt.legend(bbox_to_anchor=(1, .5), loc="center left",
                         fancybox=True, shadow=True, framealpha=0.8)
    else:
        lgd = None

    # Set labels at the center of bars
    if labels:
        ax.set_xticks(np.arange(len(labels)))
        # Determine rotation and alignmnet from labels
        if max([len(x) for x in labels]) >= 3:
            ax.set_xticklabels(labels, ha="right", rotation=45)
        else:
            ax.set_xticklabels(labels, ha="center")

        # Setting the x range to avoid large spaces between the axis and bars
        plt.xlim([min(np.arange(len(labels))) - .5,
                  max(np.arange(len(labels))) + .5])

    # Invert x-axis so that higher taxa prevalence is shown in the left
    if reverse_x:
        plt.gca().invert_xaxis()

    # Generating table structure
    if table_header:
        table = [table_header]
    else:
        table = []

    for l, x in zip(labels, data[0]):
        table.append([l, x])

    return fig, lgd, table


@set_props
def multi_bar_plot(data, labels=None, lgd_list=None, title=None,
                   ax_names=None):
    """Creates a multiple bar plot.

    Creates a multiple bar plot from a multi-dimensional array. The bar
    plots from each array will be grouped and displayed side by side.

    Parameters
    ----------
    data : numpy.array
        Single or multi-dimensional array.
    labels : list
        List of xtick labels. Should have the same length as `data`.
    title : str
        Title of the plot.
    lgd_list : list
        For categorical plots, provide the label of each category.
    ax_names : list
        List with the labels for the x-axis (first element) and y-axis
        (second element).

    Returns
    -------
    fig : matplotlib.Figure
        Figure object of the plot.
    lgd : matplotlib.Legend
        Legend object of the plot.
    table : list
        Table data in list format. Each item in the list corresponds to a
        table row.
    """

    plt.rcParams["figure.figsize"] = (8, 6)

    # Use ggpot style
    plt.style.use("ggplot")

    fig, ax = plt.subplots()

    # Set margin between bar groups
    margin = 0.1

    # Determine bar group width according to the number of data lists
    w = (1. - 2. * margin) / len(data)
    # Determine max plt height to calculate text label space
    max_height = max((x for y in data for x in y))

    # Create bar plots
    for i, d in enumerate(data):
        # Get color from 10 color list. If more than 10 colors are required,
        # randomly generate new ones
        try:
            clr = clr_list[i]
        except IndexError:
            clr = np.random.rand(3, 1)

            # Determine position of xdata
        xdata = np.arange(len(d)) + margin + (i * w)
        bplt = ax.bar(xdata, d, w, color=clr, label=lgd_list[i])

    # Add labels
    if labels:
        ax.set_xticks(np.arange(len(labels)) + 0.5)
        ax.set_xticklabels(labels, rotation=45, ha="center", fontweight="bold")

    # Set label colors
    ax.tick_params(axis="x", colors=[.4, .4, .4])
    ax.tick_params(axis="y", colors=[.3, .3, .3])

    # Add legend
    lgd = plt.legend(bbox_to_anchor=(1, .5), loc="center left",
                     fancybox=True, shadow=True, framealpha=0.8)

    # Add text to the top of bar with the corresponding value
    for b in bplt:

        # If the bar width is large enough to accommodate the text with
        # horizontal orientation
        if w >= 0.8:
            xpos = b.get_x() + b.get_width() / 2
            ypos = b.get_height() * .90
            ax.text(xpos, ypos, b.get_height(), horizontalalignment="center",
                    weight='bold', color="white", size=14)
        else:
            if b.get_height() / max_height >= 0.2:
                xpos = b.get_x() + b.get_width() / 2 * 1.3
                ypos = b.get_height() * .95
                ax.text(xpos, ypos, b.get_height(),
                        horizontalalignment="center", color="white",
                        verticalalignment="top", weight='bold',
                        size=14, rotation=90)
            else:
                xpos = b.get_x() + b.get_width() / 2
                ypos = b.get_height() * 1.1
                ax.text(xpos, ypos, b.get_height(),
                        horizontalalignment="center", color="black",
                        verticalalignment="bottom", weight='bold',
                        size=14, rotation=90)

    # Automatically adjust figure size to accommodate all labels
    # plt.tight_layout()

    return fig, lgd


@set_props
def interpolation_plot(data, title=None, ax_names=None):
    """Creates black and white interpolation plot

    Creates a black and white interpolation plot from data, which must consist
    of a 0/1 matrix for absence/presence of taxa in genes.

    Parameters
    ----------
    data : numpy.array
        Single or multi-dimensional array with plot data.
    title : str
        Title of the plot.
    ax_names : list
        List with the labels for the x-axis (first element) and y-axis
        (second element).

    Returns
    -------
    fig : matplotlib.pyplot.Figure
        Figure object of the plot.
    _ : None
        Placeholder for the legend object. Not used here but assures
        consistency across other plotting methods.
    _ : None
        Placeholder for the table header list. Not used here but assures
        consistency across other plotting methods.
    """

    plt.rcParams["figure.figsize"] = (8, 6)

    # Use ggpot style
    plt.style.use("ggplot")

    fig, ax = plt.subplots()

    # Setting the aspect ratio proportional to the data
    ar = float(len(data[0])) / float(len(data)) * .2

    ax.imshow(data, interpolation="none", cmap="Greys", aspect=ar)

    ax.grid(False)

    return fig, None, None


@set_props
def stacked_bar_plot(data, labels, legend=None, table_header=None, title=None,
                     ax_names=None, normalize=False, normalize_factor=None):
    """Creates a stacked bar plot.

    Parameters
    ----------
    data : numpy.array
        Multi-dimensional array.
    labels : list
        List of xtick labels. Should have the same length as `data`.
    title : str
        Title of the plot.
    table_header : list
        List with the header of the table object. Each element represents
        a column.
    title : str
        Title of the plot.
    ax_names : list
        List with the labels for the x-axis (first element) and y-axis
        (second element).
    normalize : bool
        If True, values of the `data` array will be normalized by the
        `normalize_factor`
    normalize_factor : int or float
        Number used to normalize values of the `data` array.

    Returns
    -------
    fig : matplotlib.Figure
        Figure object of the plot.
    lgd : matplotlib.Legend
        Legend object of the plot.
    table : list
        Table data in list format. Each item in the list corresponds to a
        table row.
    """

    plt.rcParams["figure.figsize"] = (8, 6)

    if normalize:

        data_original = np.copy(data)
        data = np.array([[y / normalize_factor for y in x] for x in data])

    if len(labels) > 10:
        plt.rcParams["figure.figsize"] = (len(labels) / 3, 6)
    else:
        plt.rcParams["figure.figsize"] = (8, 6)

    # Use ggpot style
    plt.style.use("ggplot")

    fig, ax = plt.subplots()

    # Get bottom positions for stacked bar
    bottoms = np.cumsum(np.vstack((np.zeros(data.shape[1]), data)),
                        axis=0)[:-1]

    w = .8

    # Get x positions
    xpos = [x + (w / 2) for x in range(len(labels))]

    for c, d in enumerate(data):

        if len(data) <= 10:
            c1 = c if c < 9 else c - 10
            clr = cm.Vega10(c1, 1)
        else:
            c1 = c if c < 19 else c - 20
            clr = cm.Vega20c(c1, 1)

        if c == 0:
            bplot = ax.bar(xpos, d, w, color=clr, label=legend[c], alpha=.9)
        else:
            bplot = ax.bar(xpos, d, w, color=clr, label=legend[c], alpha=.9,
                           bottom=bottoms[c])

    # Set x labels
    plt.xticks([x + (w / 2) for x in xpos], labels, ha="right", rotation=45)

    # Set legend
    if legend:
        if len(legend) <= 4:
            cols = 1
        else:
            cols = len(legend) if len(legend) < 3 else 3
        borderpad = cols * -6
        lgd = plt.legend(loc=7, fancybox=True,
                         shadow=True, framealpha=.8, ncol=cols,
                         borderaxespad=borderpad)

    # Generate table structure
    if table_header:
        table = [table_header]
    else:
        table = []
    if normalize:
        for i, lbl in enumerate(labels):
            table.append([lbl] + list(chain.from_iterable((int(x[i]), y[i])
                                      for x, y in zip(*[data_original, data]))))
    else:
        for i, lbl in enumerate(labels):
            table.append([lbl] + [int(x[i]) for x in data])

    return fig, lgd, table


@set_props
def box_plot(data, labels=None, title=None, ax_names=None):
    """Creates box (whisker) plot.

    Parameters
    ----------
    data : numpy.array
        Single or multi-dimensional array with plot data.
    labels : list
        List of xtick labels. Should have the same length as `data`.
    title : str
        Title of the plot.
    ax_names : list
        List with the labels for the x-axis (first element) and y-axis
        (second element).

    Returns
    -------
    fig : matplotlib.pyplot.Figure
        Figure object of the plot.
    _ : None
        Placeholder for the legend object. Not used here but assures
        consistency across other plotting methods.
    table : list
        Table data in list format. Each item in the list corresponds to a
        table row.
    """

    plt.rcParams["figure.figsize"] = (8, 6)

    if len(labels) > 10:
        plt.rcParams["figure.figsize"] = (len(labels) / 3, 6)
    else:
        plt.rcParams["figure.figsize"] = (8, 6)

    # Use ggpot style
    plt.style.use("ggplot")

    fig, ax = plt.subplots()

    bplot = ax.boxplot(data, patch_artist=True)

    # Change outline color, fill color and line width of the boxes
    for box in bplot["boxes"]:
        box.set(color="#7570b3", linewidth=2)
        box.set(facecolor="#1b9e77")

    # Change the color and line width of the whiskers
    for whisker in bplot["whiskers"]:
        whisker.set(color="#7570b3", linewidth=2)

    # Change color and line width of the caps
    for cap in bplot["caps"]:
        cap.set(color="#7570b3", linewidth=2)

    # Change color and line width of the medias
    for media in bplot["medians"]:
        media.set(color="#b2df8a", linewidth=2)

    # Change the style of the fliers and their fill
    for flier in bplot["fliers"]:
        flier.set(marker="o", color="#e7298a", alpha=0.5)

    if labels:
        ax.set_xticklabels(labels, rotation=45, ha="right")

    # Generate table structure
    table = [["", "1st Quartile", "Median", "3rd Quartile"]]

    for x, d in zip(labels, data):
        table.append([x, np.percentile(d, 25), np.median(d),
                      np.percentile(d, 75)])

    return fig, None, table


def histogram_smooth(data, ax_names=None, table_header=None, title=None,
                     legend=None):
    """Creates a smooth histogram-like plot.

    Creates a smooth line plot with colored areas with the same distribution
    as an histogram. It supports a multi-dimensional array, in which case
    each array will be used to create vertically stacked subplots.

    Parameters
    ----------
    data : numpy.array
        Array with plot data.
    ax_names : list
        List with the labels for the x-axis (first element) and y-axis
        (second element).
    table_header : list
        List with the header of the table object. Each element represents
        a column.
    title : str
        Title of the plot.
    legend : list
        If using a multi-dimensional array, provide the name of each
        subplot.

    Returns
    -------
    fig : matplotlib.pyplot.Figure
        Figure object of the plot.
    _ : None
        Placeholder for the legend object. Not used here but assures
        consistency across other plotting methods.
    table : list
        Table data in list format. Each item in the list corresponds to a
        table row.
    """

    from scipy.interpolate import UnivariateSpline

    plt.rcParams["figure.figsize"] = (8, 6)

    plt.style.use("ggplot")

    fig, ax = plt.subplots(nrows=3, ncols=1, sharex=True, sharey=True)

    # Start building table
    if table_header:
        table = [table_header]
    else:
        table = []

    for (p, d), axn, l in zip(enumerate(data), ax, legend):
        # Convert data into histogram
        y, x = np.histogram(d, 50)
        x = x[:-1] + (x[1] - x[0]) / 2
        # Fit a one-dimensional smoothing spline to git a given set of points
        # from the histogram
        f = UnivariateSpline(x, y, s=200)
        # Set title for subplot
        axn.set_title(l, color=clr_list[p], fontweight="bold")
        # Create subplot
        axn.plot(x, f(x), color=clr_list[p])
        # Color area below line lot
        axn.fill_between(x, f(x), alpha=0.5, color=clr_list[p])

    # If axis names are provided, add them to figure
    if ax_names:
        fig.text(0.5, 0.02, ax_names[0], ha="center", fontweight="bold",
                 color="grey", fontsize=14)
        fig.text(0.04, 0.5, ax_names[1], va="center", rotation="vertical",
                 fontweight='bold', color="grey", fontsize=14)

    # Build histogram with fixed bin range for table construction
    hist_data = []
    for d in data:
        y, x = np.histogram(d, 100, (0., 1.))
        hist_data.append(y)

    # Populate table
    for i, d in zip(np.arange(0, 1, 0.01), zip(*hist_data)):
        table.append([i] + list(d))

    return fig, None, table


@set_props
def histogram_plot(data, title=None, ax_names=None, table_header=None,
                   real_bin_num=False):
    """Creates an histogram from data.

    Parameters
    ----------
    data : numpy.array
        Array with plot data.
    title : str
        Title of the plot.
    ax_names : list
        List with the labels for the x-axis (first element) and y-axis
        (second element).
    table_header : list
        List with the header of the table object. Each element represents
        a column.
    real_bin_num : bool
        If True, then the table data will be forced to be in real numbers.

    Returns
    -------
    fig : matplotlib.Figure
        Figure object of the plot.
    lgd : matplotlib.Legend
        Legend object of the plot.
    table : list
        Table data in list format. Each item in the list corresponds to a
        table row.
    """

    # Use ggpot style
    plt.style.use("ggplot")

    plt.rcParams["figure.figsize"] = (8, 6)

    fig, ax = plt.subplots()

    c_data = Counter(data)
    max_data = max(c_data)
    min_data = min(c_data)
    if len(c_data) > 50:
        bins = int(len(c_data) / 10)
    else:
        bins = len(c_data)

    vals, b, _ = plt.hist(data, bins, histtype="stepfilled",
                                color=clr_list[0])

    plt.axvline(np.mean(data), linewidth=2, color="r", alpha=.8,
                linestyle="--")

    # Add cutom artist for legend
    mean_artist = plt.Line2D((0, 1), (0, 1), color="r", linestyle="--")

    lgd = ax.legend([mean_artist], ["Mean"], loc=0, frameon=True, fancybox=True,
                    shadow=True, framealpha=.8, fontsize="large")
    lgd.get_frame().set_facecolor("white")

    # Generate table structure
    if table_header:
        table = [table_header]
    else:
        table = []

    if real_bin_num:
        # If real_bin_num was set, then the bins in the table should be real
        # numbers
        vals, b = np.histogram(data, bins, (min_data, max_data + 1))

        c = 0
        for b, v in zip(b, vals):
            table.append(["{} - {}".format(int(c), int(b)), v])
            c = b + 1

        table.append(["{} - {}".format(int(c), max_data), vals[-1]])

    else:
        for p, val in zip(b, vals):
            table.append([p, val])

    return fig, lgd, table


@set_props
def triangular_heat(data, labels, color_label=None, title=None):
    """Creates a triangular heatmap plot.

    Parameters
    ----------
    data : numpy.array
        Triangular array with plot data.
    labels : list
        List of xtick labels. Should have the same length as `data`.
    title : str
        Title of the plot.
    color_label : str
        Label for colorbar.

    Returns
    -------
    fig : matplotlib.pyplot.Figure
        Figure object of the plot.
    _ : None
        Placeholder for the legend object. Not used here but assures
        consistency across other plotting methods.
    table : list
        Table data in list format. Each item in the list corresponds to a
        table row.
    """

    plt.style.use("ggplot")

    plt.rcParams["figure.figsize"] = (8, 6)

    fig, ax = plt.subplots()

    cmap = cm.get_cmap("jet", 100)
    cmap.set_bad("w")

    heat = ax.imshow(data, interpolation="nearest", cmap=cmap)

    # Set ticks for label positioning
    ax.set_xticks(np.arange(data.shape[1]), minor=False)
    ax.set_yticks(np.arange(data.shape[0]), minor=False)
    # Set minor ticks that will make the grid
    ax.set_xticks(np.arange(data.shape[1]) + .5, minor=True)
    ax.set_yticks(np.arange(data.shape[0]) - .5, minor=True)
    # Set x axis labels on top and y axis labels on right
    ax.xaxis.tick_top()
    # Set axis labels
    # Remove first entry of xlabel
    xlabel = ["" if x == 0 else labels[x] for x in range(len(labels))]
    ax.set_xticklabels(xlabel, rotation=45, ha="left")
    # Remove last entry of ylabel
    ylabel = ["" if x == len(labels) - 1 else labels[x] for x in
              range(len(labels))]
    ax.set_yticklabels(ylabel)
    # Remove major ticks
    plt.tick_params(axis="both", which="major", top="off", right="off",
                    left="off", labelsize=5)
    # Remove minor ticks
    plt.tick_params(axis="y", which="minor", right="off")

    plt.grid(True, which="minor")
    plt.grid(False, which="major")

    plt.gca().invert_xaxis()

    cbar = plt.colorbar(heat)
    if not color_label:
        cbar.set_label("Similarity proportion")
    else:
        cbar.set_label(color_label)

    # Generate table
    table = [[""] + [x for x in labels[::-1]]]
    for p, sp in enumerate(labels):
        table.append([sp] + list(data[p])[::-1])

    return fig, None, table


@set_props
def outlier_densisty_dist(data, outliers, outliers_labels=None, ax_names=None,
                          title=None):
    """Creates a density distribution for outlier plots.

    Parameters
    ----------
    data : numpy.array
        1D array containing data points.
    outliers : numpy.array
        1D array containing the outliers.
    outliers_labels : list or numpy.array
        1D array containing the labels for each outlier.
    title : str
        Title of the plot.

    Returns
    -------
    fig : matplotlib.Figure
        Figure object of the plot.
    lgd : matplotlib.Legend
        Legend object of the plot.
    table : list
        Table data in list format. Each item in the list corresponds to a
        table row.
    """

    plt.rcParams["figure.figsize"] = (8, 6)

    fig, ax = plt.subplots()

    # Create density function
    sns.distplot(data, rug=True, hist=False, color="black")

    # Plot outliers
    ax.plot(outliers, np.zeros_like(outliers), "ro", clip_on=False,
            label="Outliers")

    # Create legend
    lgd = ax.legend(frameon=True, loc=2, fancybox=True, shadow=True,
                    framealpha=.8, prop={"weight": "bold"})

    if outliers_labels:
        table = [[os.path.basename(x)] for x in outliers_labels]
    else:
        table = None

    return fig, lgd, table


@set_props
def sliding_window(data, window_size, ax_names=None, table_header=None,
                   title=None):
    """Creates a sliding window plot.

    Parameters
    ----------
    data : numpy.array
        Array with plot data.
    window_size : int
        Length of window size for sliding window.
    table_header : list
        List with the header of the table object. Each element represents
        a column.
    title : str
        Title of the plot.
    ax_names : list
        List with the labels for the x-axis (first element) and y-axis
        (second element).

    Returns
    -------
    fig : matplotlib.pyplot.Figure
        Figure object of the plot.
    _ : None
        Placeholder for the legend object. Not used here but assures
        consistency across other plotting methods.
    table : list
        Table data in list format. Each item in the list corresponds to a
        table row.
    """

    plt.style.use("ggplot")

    plt.rcParams["figure.figsize"] = (8, 6)

    data = np.array(data)
    x = np.arange(0, len(data) * window_size, window_size)

    table_data = []

    fig, ax = plt.subplots()

    ax.set_axis_bgcolor("#f2f2f2")

    xnew = np.linspace(x.min(), x.max(), 500)
    xsmooth = spline(x, data, xnew)

    # p = ax.plot(xnew, xsmooth)

    for i in range(0, len(xsmooth)):
        ax.plot(xnew[i:i + 2], xsmooth[i:i + 2], color=cm.rainbow(1 - (xsmooth[i] / float(max(
            xsmooth))), 1), linewidth=2.5)
        table_data.append([xnew[i], xsmooth[i]])

    ax.set_ylim([min(data) - min(data) * .1, max(data) + max(data) * .1])

    if table_header:
        table = [table_header] + table_data

    else:
        table = table_data

    return fig, None, table


__author__ = "Diogo N. Silva"
__credits__ = ["Diogo N. Silva", "Tiago F. Jesus"]



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
#
#  Copyright 2012 Unknown <diogo@arch>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import warnings

# Suppress import warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore")

    import argparse
    import configparser
    import time
    import sys
    from glob import glob

    try:
        from process.sequence import *
        from base.plotter import *
        from process.base import print_col, GREEN, RED, YELLOW, CleanUp
    except ImportError:
        from trifusion.process.sequence import *
        from trifusion.base.plotter import *
        from trifusion.process.base import print_col, GREEN, RED, YELLOW,\
            CleanUp


class HandledException(Exception):
    pass


def generate_cfg_template():

    template_fh = open("stats_template.ini", "w")

    template_fh.write("""
# Configuration template file for statistical analyses that can be passed to
# TriStats.py using the -cfg option. Section and option names mimic the names
# Present in the Statistics side panel.

# For each option, three values are possible but not necessarily all used.
# The allowed values are provided above each option.
# The values are:
#   - species: Creates a plot focusing on the taxa
#   - average: Creates a plot that averages over all alignments
#   - gene: Creates a plot for a single gene. Can only be used with single
#           input files

[General Information]
# Options available: species average
distribution_sequence_size: species average
# Options available: species average
proportion_nucleotides_residues: species average
# Options available: average
distribution_taxa_frequency: average

[Polymorphism and Variation]
# Options available: gene species average
sequence_similarity: species average
# Options available: gene species average
segregating_sites: species average
# Options available: average
alignment_pol_correlation: average
# Options available: gene average
allele_frequency_spectrum: average

[Missing Data]
# Options available: average
gene_occupancy: average
# Options available: species average
distribution_missing_genes: species average
# Options available: species average
distribution_missing_data: species average
# Options available: average
cumulative_distribution_missing_genes: average

[Outlier Detection]
# Options available: species average
missing_data_outliers: species average
# Options available: species average
segregating_sites_outliers: species average
# Options available: species average
sequence_size_outliers: species average
    """)

    template_fh.close()


def main_checks(arg):

    if not arg.infile and not arg.generate_cfg:
        print_col("Must provide input data using the '-in' option", RED, 2)


def main():

    parser = argparse.ArgumentParser(description="Command line interface for "
                                                 "TriFusion Statistics module")

    # Main execution
    main_exec = parser.add_argument_group("Main execution")
    main_exec.add_argument("-in", dest="infile", nargs="+",
                           help="Provide the input files.")
    main_exec.add_argument("-o", dest="project_name",
                           help="Name of the output directory")
    main_exec.add_argument("-cfg", dest="config_file",
                           help="Name of the configuration file with the "
                                "statistical analyses to be executed")
    main_exec.add_argument("--generate-cfg", dest="generate_cfg",
                           action="store_const", const=True,
                           help="Generates a configuration template file")
    main_exec.add_argument("-quiet", dest="quiet", action="store_const",
                           const=True, default=False, help="Removes all"
                           " terminal output")

    arg = parser.parse_args()

    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)

    main_checks(arg)

    stats_main(arg)


@CleanUp
def stats_main(args):

    print_col("Executing TriStats module at %s %s" % (
        time.strftime("%d/%m/%Y"), time.strftime("%I:%M:%S")), GREEN, 2)

    if args.generate_cfg:
        print_col("Generating configuration template file", GREEN, 2)
        return generate_cfg_template()

    # Create temporary directory
    tmp_dir = ".trifusion-temp"
    if not os.path.exists(tmp_dir):
        os.makedirs(tmp_dir)

    # Set path to temporary sqlite database
    sql_db = os.path.join(tmp_dir, "trifusion.db")

    # Arguments
    input_files = args.infile
    output_dir = args.project_name
    config_file = args.config_file

    # Read configuration file
    print_col("Reading configuration file", GREEN, 2)
    settings = configparser.ConfigParser()
    settings.read(config_file)

    # Parse alignments
    # Support wildcards as arguments for windows
    fl = []
    if sys.platform in ["win32", "cygwin"]:
        for p in input_files:
            fl += glob(p)
        input_files = fl

    print_col("Parsing %s alignments" % len(input_files), GREEN, 2)
    alignments = AlignmentList(input_files, sql_db=sql_db)

    # Create output dir
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    # Variable mapping each available option with the appropriate statistics
    # and plotting methods
    func_map = {
        ("general information", "distribution_sequence_size", "species"):
            [alignments.average_seqsize_per_species,
             (box_plot, "avg_seqsize_species.png")],

        ("general information", "distribution_sequence_size", "average"):
            [alignments.average_seqsize,
             (histogram_plot, "avg_seqsize.png")],

        ("general information", "proportion_nucleotides_residues", "species"):
            [alignments.characters_proportion_per_species,
             (stacked_bar_plot, "char_proportions_sp.png")],

        ("general information", "proportion_nucleotides_residues", "average"):
            [alignments.characters_proportion,
             (bar_plot, "char_proportions.png")],

        ("general information", "distribution_taxa_frequency", "average"):
            [alignments.taxa_distribution,
             (histogram_plot, "distribution_taxa_frequency.png")],

        ("polymorphism and variation", "sequence_similarity", "species"):
            [alignments.sequence_similarity_per_species,
             (triangular_heat, "similarity_distribution_sp.png")],

        ("polymorphism and variation", "sequence_similarity", "average"):
            [alignments.sequence_similarity,
             (histogram_plot, "similarity_distribution.png")],

        ("polymorphism and variation", "sequence_similarity", "gene"):
            [alignments.sequence_similarity_gene,
             (sliding_window, "similarity_distribution_gn.png")],

        ("polymorphism and variation", "segregating_sites", "species"):
            [alignments.sequence_segregation_per_species,
             (triangular_heat, "segregating_sites_sp.png")],

        ("polymorphism and variation", "segregating_sites", "average"):
            [alignments.sequence_segregation,
             (histogram_plot, "segregating_sites.png")],

        ("polymorphism and variation", "segregating_sites", "gene"):
            [alignments.sequence_segregation_gene,
             (sliding_window, "segregating_sites_gn.png")],

        ("polymorphism and variation", "alignment_pol_correlation", "average"):
            [alignments.length_polymorphism_correlation,
             (scatter_plot, "length_polymorphism_correlation.png")],

        ("polymorphism and variation", "allele_frequency_spectrum", "average"):
            [alignments.allele_frequency_spectrum,
             (histogram_plot, "allele_frequency_spectrum.png")],

        ("polymorphism and variation", "allele_frequency_spectrum", "gene"):
            [alignments.allele_frequency_spectrum_gene,
             (histogram_plot, "allele_frequency_spectrum_gn.png")],

        ("missing data", "gene_occupancy", "average"):
            [alignments.gene_occupancy,
             (interpolation_plot, "gene_occupancy.png")],

        ("missing data", "distribution_missing_genes", "species"):
            [alignments.missing_genes_per_species,
             (bar_plot, "missing_gene_distribution.png")],

        ("missing data", "distribution_missing_genes", "average"):
            [alignments.missing_genes_average,
             (histogram_plot, "missing_gene_distribution_avg.png")],

        ("missing data", "distribution_missing_data", "species"):
            [alignments.missing_data_per_species,
             (stacked_bar_plot, "missing_data_distribution_sp.png")],

        ("missing data", "distribution_missing_data", "average"):
            [alignments.missing_data_distribution,
             (histogram_smooth, "missing_data_distribution.png")],

        ("missing data", "cumulative_distribution_missing_genes", "average"):
            [alignments.cumulative_missing_genes,
             (bar_plot, "cumulative_distribution_missing_genes.png")],

        ("outlier detection", "missing_data_outliers", "species"):
            [alignments.outlier_missing_data_sp,
             (outlier_densisty_dist, "Missing_data_outliers_sp.png")],

        ("outlier detection", "missing_data_outliers", "average"):
            [alignments.outlier_missing_data,
             (outlier_densisty_dist, "Missing_data_outliers.png")],

        ("outlier detection", "segregating_sites_outliers", "species"):
            [alignments.outlier_segregating_sp,
             (outlier_densisty_dist, "Segregating_sites_outliers_sp.png")],

        ("outlier detection", "segregating_sites_outliers", "average"):
            [alignments.outlier_segregating,
             (outlier_densisty_dist, "Segregating_sites_outliers.png")],

        ("outlier detection", "sequence_size_outliers", "species"):
            [alignments.outlier_sequence_size_sp,
             (outlier_densisty_dist, "Sequence_size_outliers_sp.png")],

        ("outlier detection", "sequence_size_outliers", "average"):
            [alignments.outlier_sequence_size,
             (outlier_densisty_dist, "Sequence_size_outliers.png")]
    }

    print_col("Parsing configuation file options", GREEN, 2)

    # Iterate over each individual option
    for section in settings.sections():
        for option, val in settings.items(section):
            for i in val.split():

                section = section.lower()
                # Check if current option is available or supported
                if (section, option, i) in func_map:
                    print_col("Generating plot for option: %s - %s - %s" %
                              (section, option, i), GREEN, 2)
                    # Get appropriate method list
                    funcs = func_map[(section, option, i)]
                    # Retrieve plot data using statistics method
                    plot_data = funcs[0]()

                    # Check for exceptions in plot data
                    if "exception" in plot_data:
                        if plot_data["exception"] is EmptyData:
                            print_col("Option %s - %s - %s has no data for "
                                      "plotting" % (section, option, i),
                                      YELLOW, 2)
                        if plot_data["exception"] is InvalidSequenceType:
                            print_col("Invalid sequence type for option %s - "
                                      "%s - %s (%s)" %
                                      (section, option, i,
                                       alignments.sequence_code[0]), YELLOW, 2)
                        continue

                    # Generate plot object
                    plot_obj, _, lgd = funcs[1][0](**plot_data)
                    plot_obj.tight_layout()

                    # Save plot to file, including the legend object, if
                    # available
                    if lgd:
                        plot_obj.savefig(join(output_dir, funcs[1][1]),
                                         bbox_extra_artists=(lgd,), dpi=200)
                    else:
                        plot_obj.savefig(join(output_dir, funcs[1][1]),
                                         dpi=200)
                else:
                    print_col("Invalid option: %s - %s - %s. Skipping." %
                              (section, option, i), YELLOW, 2)

if __name__ == '__main__':
    main()


__author__ = "Diogo N. Silva"


#!/usr/bin/env python2
# -*- coding: utf-8 -*-
#
#  Copyright 2012 Unknown <diogo@arch>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

"""
This module deals with the conversion of protein sequences into their
corresponding nucleotide sequences. Since the conversion from protein to DNA
cannot be made without knowing the nucleotide sequence, this module contains
functions that compile and store DNA sequences, convert them into amino acid
sequences and then tries to match them to the original protein sequences.
"""

try:
    from process.error_handling import KillByUser
except ImportError:
    from trifusion.process.error_handling import KillByUser

from os.path import join
import subprocess
import os

dna_map = {
    'ATA': 'I', 'ATC': 'I', 'ATT': 'I', 'ATG': 'M',
    'ACA': 'T', 'ACC': 'T', 'ACG': 'T', 'ACT': 'T',
    'AAC': 'N', 'AAT': 'N', 'AAA': 'K', 'AAG': 'K',
    'AGC': 'S', 'AGT': 'S', 'AGA': 'R', 'AGG': 'R',
    'CTA': 'L', 'CTC': 'L', 'CTG': 'L', 'CTT': 'L',
    'CCA': 'P', 'CCC': 'P', 'CCG': 'P', 'CCT': 'P',
    'CAC': 'H', 'CAT': 'H', 'CAA': 'Q', 'CAG': 'Q',
    'CGA': 'R', 'CGC': 'R', 'CGG': 'R', 'CGT': 'R',
    'GTA': 'V', 'GTC': 'V', 'GTG': 'V', 'GTT': 'V',
    'GCA': 'A', 'GCC': 'A', 'GCG': 'A', 'GCT': 'A',
    'GAC': 'D', 'GAT': 'D', 'GAA': 'E', 'GAG': 'E',
    'GGA': 'G', 'GGC': 'G', 'GGG': 'G', 'GGT': 'G',
    'TCA': 'S', 'TCC': 'S', 'TCG': 'S', 'TCT': 'S',
    'TTC': 'F', 'TTT': 'F', 'TTA': 'L', 'TTG': 'L',
    'TAC': 'Y', 'TAT': 'Y', 'TAA': '', 'TAG': '',
    'TGC': 'C', 'TGT': 'C', 'TGA': '', 'TGG': 'W'}


def translate(sequence):
    """
    Translates a DNA string into an amino acid sequence
    :param sequence: string. DNA sequence
    :return: String. Protein sequence
    """

    sequence = sequence.lower().replace("-", "")
    aa_sequence = ""

    for i in range(0, len(sequence), 3):
        codon = sequence[i:i + 3]

        # Check if codon is multiple of 3. If not, place an X (missing data)
        if len(codon) == 3:
            # Check if there is missing data on the codon. If so, place an X
            if "n" in codon:
                aa = ""
            else:
                try:
                    aa = dna_map[codon.upper()]
                except KeyError:
                    aa = ""
        else:
            aa = ""

        aa_sequence += aa

    return aa_sequence


def create_db(f_list, dest="./", ns=None):
    """
    Creates a fasta database file containing the translated protein sequences
    from the cds files. The final transcripts.fas file will be use
    by USEARCH to get matches between the original protein sequences and their
    nucleotide counterparts. A dictionary database will also be created where
    the transcript headers will be associated with the original DNA sequence,
    so that they will be later retrieved
    :param f_list. List, containing the file names of the transcript files
    """

    output_handle = open(join(dest, "transcripts.fas"), "w")
    id_dic = {}

    if ns:
        if ns.stop:
            raise KillByUser("")

        ns.progress = 0
        ns.max_pb = len(f_list)

    for f in f_list:
        handle = open(f)
        seq = ""
        header = ""

        if ns:
            if ns.stop:
                raise KillByUser("")
            ns.progress += 1

        for line in handle:

            if ns:
                if ns.stop:
                    raise KillByUser("")

            if line.startswith(">"):
                if seq != "":
                    aa_seq = translate(seq)
                    output_handle.write(">%s\n%s\n" % (header, aa_seq))
                    id_dic[header] = seq

                header = line.strip()[1:].replace(" ", ";;")
                seq = ""
            else:
                seq += line.strip()

    output_handle.close()

    return id_dic


def create_query(input_list, dest="./"):
    """
    To speed things up, all sequences in the input protein files will be
    concatenated into a single file, which will be used as query in USEARCH.
    :param input_list: List, with file names of the protein files to convert
    """

    f_handle = open(join(dest, "query.fas"), "w")
    query_db = {}

    for f in input_list:
        handle = open(f)
        for line in handle:
            if line.startswith(">"):
                query_db[f].append(line.strip()[1:].replace(" ", ";;"))
                f_handle.write(line.replace(" ", ";;"))
            else:
                f_handle.write(line)

        handle.close()

    f_handle.close()

    return query_db


def create_query_from_dict(protein_dict):
    """
    Analogous to create_query, but begins from a dictionary provided by
    processed group files
    :param protein_dict: dictionary
    """

    query_db = {"group": []}
    f_handle = open("query", "w")

    for cl in protein_dict:
        for seq_id, seq in cl:
            query_db[cl].append(seq_id.replace(" ", ";;"))
            f_handle.write(">%s\n%s\n" % (seq_id.replace(" ", ";;"), seq))

    f_handle.close()

    return query_db


def pair_search(usearch_bin, dest="./"):
    """
    This will use USEARCH to search for translated transcript sequences
    identical to the original protein files
    """

    query_path = join(dest, "query.fas")
    db_path = join(dest, "transcripts.fas")
    out_path = join(dest, "pairs.out")

    subprocess.Popen([usearch_bin,
                      "-usearch_global",
                      query_path,
                      "-db",
                      db_path,
                      "-id",
                      "1",
                      "-maxaccepts",
                      ".9",
                      "-blast6out",
                      out_path]).wait()


def get_pairs(dest="./", ns=None):
    """
    Parses the output of USEARCH and creates a dictionary with the header
    pairs between original protein and transcripts
    """

    file_h = open(join(dest, "pairs.out"))
    pair_db = {}

    if ns:
        if ns.stop:
            raise KillByUser("")
        p = 0
        with open(join(dest, "pairs.out")) as f:
            for p, _ in enumerate(f):
                pass
        ns.max_pb = p + 1
        ns.progress = 0

    for l in file_h:

        if ns:
            if ns.stop:
                raise KillByUser("")
            ns.progress += 1

        fields = l.split("\t")
        pair_db[fields[0]] = fields[1]

    file_h.close()
    return pair_db


def convert_protein_file(pairs, group_obj, id_db, output_dir, shared_ns):
    """
    A given protein file will be converted into their corresponding nucleotide
    sequences using a previously set database using the create_db function
    :return:
    """

    # Create handle for file storing bad sequence headers.
    bad_file = open(join(output_dir, "missed_sequences.log"), "w")

    for line, cl in zip(group_obj.groups(),
                        group_obj.iter_species_frequency()):

        if shared_ns:
            if shared_ns.stop:
                raise KillByUser("")

        if group_obj._get_compliance(cl) == (1, 1):

            line = group_obj._remove_tx(line)

            fields = line.split(":")
            orto_name = fields[0]
            seq_headers = fields[-1].split()

            f_handle = open(join(output_dir, orto_name) + ".fas", "w")

            for h in seq_headers:
                if h in pairs:
                    seq = id_db[pairs[h]]
                    shared_ns.good += 1
                    f_handle.write(">%s\n%s\n" % (h.replace(";;", " "), seq))
                else:
                    shared_ns.missed += 1
                    bad_file.write("{}\t{}\n".format(orto_name, h))


def convert_group(sqldb, cds_file_list, protein_db, group_sequences,
                usearch_bin, output_dir, shared_namespace=None):
    """
    Convenience function that wraps all required operations to convert protein
    to nucleotide files from a Group object
    """

    if shared_namespace:
        shared_namespace.act = "Creating database"
        shared_namespace.missed = 0
        shared_namespace.good = 0
    # Create database
    id_db = create_db(cds_file_list, output_dir, shared_namespace)

    if shared_namespace:
        shared_namespace.act = "Creating query"

        # Kill switch
        if shared_namespace.stop:
            raise KillByUser("")

    # Create query for USEARCH
    group_sequences.retrieve_sequences(sqldb, protein_db, output_dir,
                                       outfile="query.fas",
                                       shared_namespace=shared_namespace)

    if shared_namespace:
        # Kill switch
        if shared_namespace.stop:
            raise KillByUser("")

    # Execute search
    if shared_namespace:
        shared_namespace.act = "Performing search"
    pair_search(usearch_bin, output_dir)

    if shared_namespace:
        # Kill switch
        if shared_namespace.stop:
            raise KillByUser("")

    pair_db = get_pairs(output_dir, ns=shared_namespace)
    # Convert files

    if shared_namespace:
        shared_namespace.act = "Converting to nucleotide"
    convert_protein_file(pair_db, group_sequences, id_db, output_dir,
                         shared_namespace)

    # Remove temporary files
    temp_files = [join(output_dir, "query.fas"),
                  join(output_dir, "transcripts.fas"),
                  join(output_dir, "pairs.out")]

    for f in temp_files:
        os.remove(f)

__author__ = "Diogo N. Silva"



#!/usr/bin/python2
# -*- coding: utf-8 -*-

import sqlite3 as lite
import os


def execute(db_dir, similar_seqs_file):
    con = lite.connect(os.path.join(db_dir, "orthoDB.db"))

    with con:

        cur = con.cursor()

        file_handle = open(similar_seqs_file)

        for line in file_handle:
            if line.strip() != "" and line:
                f = line.split("\t")

                l = (f[0], f[1], f[2], f[3], float(f[4]), float(f[5]),
                    float(f[6]), float(f[7]))
                
                cur.execute("INSERT INTO SimilarSequences VALUES(?, ?, ?, ?, "
                            "?, ? ,?, ?)", l)


if __name__ == "__main__":
    execute(".", "sss_nodups.txt")

__author__ = "Fernando Alves and Diogo N. Silva"
